from ..formula.kendrick_mass_defect import kendrick_mass_defect
from matplotlib import pyplot as plt
def kendrick_mass_defect_plot(msTuple, base = 'CH2', rounding = 'even', **kwargs):
    """ 
	Docstring for function PyKrev.kendrick_mass_defect_plot.py
	====================
	This function takes an msTuple and creates a kendrick mass defect plot.
    
	Use
	----
	kendrick_mass_defect_plot(Y)
    
	Returns the figure and axes handles and a tuple containing two numpy arrays, the first contains the kendrick mass and the second the kendrick mass defect. 
    
	Parameters
	----------
	Y: msTuple
	Base: Atom group used to define the Kendrick mass.

    Rounding: One of 'ceil', 'floor', or 'even', see pk.kendrickMass()

    **kwargs: key word arguments for pyplot.scatter(). 

    Info
	----------
    Calculation taken from Hughey et al (2001) 
    "Kendrick Mass Defect Spectrum: A Compact Visual Analysis for Ultrahigh-Resolution Broadband Mass Spectra"
    Note: Rounding calclations may lead to artefacts in complex datasets with many peaks. 
    We recommend experimenting with different rounding methods when making kmd plots.
    """
    kendrickMass, kendrickMassDefect = kendrick_mass_defect(msTuple,base=base,rounding=rounding)
    plt.scatter(kendrickMass,kendrickMassDefect, **kwargs)
    plt.xlabel('Kendrick Mass')
    plt.ylabel('Kendrick Mass Defect')
    fig = plt.gcf()
    ax = plt.gca()
    return fig, ax, (kendrickMass, kendrickMassDefect)
    


from .aromaticity_index import aromaticity_index 
from .double_bond_equivalent import double_bond_equivalent 
from .element_counts import element_counts
from .element_ratios import element_ratios
from .nominal_oxidation_state import nominal_oxidation_state
from .find_intersections import find_intersections
from .read_formularity import read_formularity
from .read_batch_formularity import read_batch_formularity
from .calculate_mass import calculate_mass
from .kendrick_mass_defect import kendrick_mass_defect
from .filter_spectral_interference import filter_spectral_interference
from .read_corems import read_corems
from .msTuple import msTuple
from .msTupleDict import msTupleDict
from .average_mstuple import average_mstuple
from .read_csv import read_csv


import pandas as pd
import numpy as np 
from .msTuple import msTuple
def read_formularity(report_name,pi_col = []):
    """ 
	Docstring for function PyKrev.read_formularity
	====================
	This function reads the report csv file produced by formularity software and returns an msTuple.
    
	Use
	----
	read_formularity(report_name)
    
	Returns an msTuple. 

	Parameters
	----------
	report_name: name of csv file that the formularity report to be read is saved as. 
	pi_col: name of the column in that file that peak intensities are found in. If not given the last column is used. 

    Note: PyKrev will filter out formula with 13C assignments
    """
    report = pd.read_csv(report_name)
    notIsotopologue = report['C13'] == 0 # boolean array of only non isotopologues
    report = report[notIsotopologue]
    report.reset_index(drop = True, inplace = True) #reset the index to account for the removed isotopologues
    C = report['C']
    H = report['H']
    O = report['O']
    N = report['N'] 
    S = report['S']
    P = report['P']
    if not pi_col: 
        I = report.iloc[:,-1] #take the final column of the report file to contain peak intensities. Not sure how stable this is.
    else:
        I = report[pi_col] #user supplied a column name for the peak intensities.
    MZ = report['Mass']
    G = report['Class']
    ME = report['Error_ppm']
    molecular_formula = []
    mass_charge = np.array([])
    mass_error = np.array([])
    peak_intensities = np.array([])
    compound_class = []
    for n in range(0,len(C)):
        if C[n] == 0 and H[n] == 0: #If there isn't a count for C and H don't include the formula
            pass
        else:
            temp_formula = 'C'+str(C[n])+'H'+str(H[n]) #add C and H then ... 
            if N[n] > 0:
                temp_formula = temp_formula + 'N' + str(N[n])
            if O[n] > 0:
                temp_formula = temp_formula + 'O' + str(O[n])
            if P[n] > 0:
                temp_formula = temp_formula + 'P' + str(P[n])
            if S[n] > 0:
                temp_formula = temp_formula + 'S' + str(S[n])
            molecular_formula.append(temp_formula)
            peak_intensities = np.append(peak_intensities,I[n])
            mass_charge =  np.append(mass_charge,MZ[n])
            mass_error = np.append(mass_error,ME[n])
            compound_class.append(G[n])
    return msTuple(molecular_formula, peak_intensities, mass_charge) #return an msTuple


from matplotlib import pyplot as plt
from scipy.stats import binned_statistic_2d
from ..formula.element_ratios import element_ratios 
from ..formula.element_counts import element_counts


def van_krevelen_histogram (msTuple, x_ratio = 'OC', y_ratio ='HC', **kwargs): 
    """ 
	Docstring for function PyKrev.van_krevelen_histogram
	====================
	This function takes an msTuple and plots a van Krevelen histogram. 
    
	Use
	----
	van_krevelen_histogram(Y)
    
	Returns the figure and axes handles and a density index score if bins are provided as integers (see PyKrev.density_index).   
    
	Parameters
	----------
	Y: msTuple
    x_ratio: string, element ratio to plot on x axis, given numerator denominator e.g. 'OC'
    y_ratio: string, element ratio to plot on y axis, given numerator denominator e.g. 'HC'
	**kwargs for pyplot.hist2d() See: https://matplotlib.org/api/_as_gen/matplotlib.pyplot.hist2d.html.
    """
    ratio_list = element_ratios(msTuple,ratios=[x_ratio,y_ratio])
    if 'bins' not in kwargs: 
        kwargs['bins'] = 20
        xbins = 20
        ybins = 20 
        d_index = density_index(ratio_list,xbins,ybins)
    elif isinstance(kwargs['bins'],int):
        xbins = kwargs['bins']
        ybins = kwargs['bins']
        d_index = density_index(ratio_list,xbins,ybins)
    elif isinstance(kwargs['bins'][0],int) and isinstance(kwargs['bins'][1],int):
        xbins = kwargs['bins'][0]
        ybins = kwargs['bins'][1]
        d_index = density_index(ratio_list,xbins,ybins)
    else: d_index = None
    x_axis = []
    y_axis = []
    for ratios in ratio_list:
        x_axis.append(ratios['OC'])
        y_axis.append(ratios['HC'])
    plt.hist2d(x_axis,y_axis,**kwargs)
    plt.xlabel('Atomic ratio of O/C')
    plt.ylabel('Atomic ratio of H/C')
    fig = plt.gcf()
    ax = plt.gca()
    return fig, ax, d_index
        
    
def density_index (ratio_list,xbins=20,ybins=20):
    
    """      
	Docstring for function PyKrev.van_krevelen_histogram
	====================
	This function takes a list of H/C-O/C ratios and calculates a density score. 
	This is achieved by dividing the average number of points by the number of points in the most populated. 
	Giving average relative density. For 100 bins a score of 1 means all bins are equally dispersed, and a score of 0.01 
	means all points fall into one bin.
    
	Use
	----
	density_index(Y)
    
	Returns a float containing density index score.   
    
	Parameters
	----------
	Y: A list of atom ratios (must contain H/C and O/C). See PyKrev.element_ratios.
    """
    x = []
    y = []
    for ratios in ratio_list:
        x.append(ratios['OC'])
        y.append(ratios['HC'])
    #count the number of points that fall into each of our pre-defined bins
    bin_results = binned_statistic_2d(x,y,None,'count',bins = (xbins,ybins))
    bin_counts = bin_results.statistic
    #compute the average bin density
    average_bin_density = bin_counts.mean()
    #compute the max bin density 
    max_bin_density = bin_counts.max()
    #compute the density distribution 
    d_index = average_bin_density/max_bin_density   
    return d_index


import numpy as np
from .calculate_mass import calculate_mass
from .msTuple import msTuple
def average_mstuple(Y, intensityMethod = 'mean', mzMethod = 'mean', minOccurrence = 1, zeroValues = True, stdDev = False):
    """ 
    Docstring for function pykrev.average_mstuple
    ==========
    Averages all the spectra in an msTupleDict and returns the average msTuple
    
    Use 
    ----------
    average_mstuple(Y) 
    
    Returns an msTuple with the spectra in the msTuple dictionary averaged according to their intensity and mz. 
    
    Parameters 
    ----------
    Y: An msTupleDict
    
    Give a description of all possible input parameters and their type, starting with positional arguments. 

    intensityMethod: string, one of: 
        'mean' - use the mean intensity for a given formula 
        'max' - use the max intensity for a given formula 
        'sum' - use the sum intensity for a given forumla
    
    mzMethod: string, one of:
        'mean' - use the mean mz for each formula 
        'median' - use the median mz for each formula
        'monoisotopic' - use the monoisotopic mz for each formula 

    minOccurance: int, the minimum number of samples a formula must appear in to be included in the intensity averaging

    zeroValues: bool, include zero (i.e. missing formula) values in average intensity calculation

    stdDev: boolean, if true return also a tuple containing 
                    (i)  array of the sample tandard deviations of each formula intensity in the output msTuple, 
                    (ii) array of the standard deviatons of each mz in the output msTuple

    """
    #Tests
    Y.validate()
    assert intensityMethod in ['mean','max','sum'], "You must provide a valid method"
    assert mzMethod in ['mean','median', 'monoisotopic'], "You must provide a valid method"
    assert minOccurrence > 0, "minOccurrence must be at least 1"
    #Setup
    OrdinationMat = Y.to_OrdinationMatrix(impute_value = 0)
    row,col = OrdinationMat.shape
    formulaNames = OrdinationMat.columns
    outputFormula = []
    outputIntensity = []
    outputMZ  = []
    stdDevIntensity = []
    stdDevMZ = []
    #Main
    for i in range(col):
        ##Test whether Occurrence matches min Occurrence
        nonZeroIdx = OrdinationMat.iloc[:,i].to_numpy().nonzero()[0]
        occurrence = len(nonZeroIdx)
        if occurrence >= minOccurrence:
            ## Remove or keep zero values
            if zeroValues == True:
                testArray = OrdinationMat.iloc[:,i].to_numpy()
            elif zeroValues == False:
                testArray = OrdinationMat.iloc[:,i].to_numpy()[nonZeroIdx]
            ## Perform the averaging
            if intensityMethod == 'mean':
                outputIntensity.append(testArray.mean())
            elif intensityMethod == 'max':
                outputIntensity.append(testArray.max())                
            elif intensityMethod == 'sum':
                outputIntensity.append(testArray.sum())
            formula = formulaNames[i]
            outputFormula.append(formula)
            ## Determine the mz across spectra where this formula was found 
            if mzMethod == 'monoisotopic':
                outputMZ.append(calculate_mass(formula))
            mzArray = np.array([])
            for v in Y.values():
                try:
                    idx = v.formula.index(formula)
                except ValueError:
                    continue
                mzArray = np.append(mzArray,v.mz[idx])
            if mzMethod == 'mean':
                outputMZ.append(mzArray.mean())
            elif mzMethod == 'median':
                outputMZ.append(mzArray.median())
            ## Determine the standard deviations
            if stdDev == True:
                stdDevIntensity.append(testArray.std(ddof = 1))
                stdDevMZ.append(mzArray.std(ddof = 1))
    assert len(outputIntensity) == len(outputMZ) == len(outputFormula)
    outputMZ = np.array(outputMZ)
    outputIntensity = np.array(outputIntensity)
    if stdDev == True:
        assert len(outputIntensity) == len(stdDevIntensity)
        return msTuple(outputFormula, outputIntensity, outputMZ), (stdDevIntensity, stdDevMZ)
    else:
        return msTuple(outputFormula, outputIntensity, outputMZ)




            






from .element_counts import element_counts
import numpy as np
def double_bond_equivalent(msTuple):
    """ 
	Docstring for function pyKrev.double_bond_equivalent
	==========
	This function takes an msTuple and returns the double bond equivalent of each formula in the formula list.
    
	Use
	----------
	double_bond_equivalent(Y)
    
	Returns a numpy array of len(Y[0]) in which each item , i , is the double bond equivalent of Y[0][i].  
    
	Parameters
	----------
	Y: msTuple OR a list of molecular formula strings
    
	Info
	----------
	Double bond equivalent (DBE; UN; degree of unsaturation; PBoR [Pi Bonds or Rings]): 
	The number of molecules of H2 that would have to be added to a molecule to convert all pi bonds to single bonds, 
	and all rings to acyclic structures. 
	The DBE number can be calculated from the formula using the following equation:
	DBE = UN = PBoR = C - (H/2) + (N/2) +1,
	where: C = number of carbon atoms, H = number of hydrogen and halogen atoms, and N = number of nitrogen atoms.
    """    
    #Setup
    count_list = element_counts(msTuple)
    DBE_array = np.array([])
    #Main
    for count in count_list:
        Halogens = ['H','Cl','Br','I','F','At','Ts']
        Hal = 0 
        for el in Halogens: 
            try:
                Hal += count[el]
            except KeyError:
                pass 
            
        DBE_counts = count['C'] - (Hal/2) + (count['N']/2) + 1 
        DBE_array = np.append(DBE_array,DBE_counts)        
    return DBE_array


import numpy as np
import pandas as pd
def normalise_intensity(input_matrix, norm_method = 'sum', norm_subset =  'ALL', p_L = 500, p_P = 0.5, norm_transform = 'none', norm_direction = 'rows'):
    """ 
    Docstring for function pykrev.normalise_intensity 
    ====================
    This function takes an intensity data matrix and applies a normalisation method on the rows (i.e samples) or columns (i.e metabolites) of the data.
    Normalisation consists of two processes: (1) applying a subset method to the data, (2) generating normalisation factors (e.g. mean, median, zscore)
    based on that subset which are then applied to the entire dataset. Note: if you are normalising column wise, you cannot subset the data. 

    Use
    ----------
    normalise_intensity(Y)
    
    Returns a numpy array or pd.dataframe of shape(Y) in which each value corresponds to the row or column normalised intensity.  
    
    Parameters
    ----------
    Y: A numpy array of shape Y[samples,formula] containing peak intensities OR an ordination matrix produced by pk.ordination matrix 
    norm_method: A string decribing the relative intensity metric to be used. One of:
            - 'center' : center the data by subtracting the mean i.e. mean(Y[i,:]) == 0
            - 'zscore': zscore normalisation i.e. mean(Y[i,:]) == 0, sd(Y[i,:]) == 1
            - 'pareto': pareto normalisation i.e. mean(Y[i,:]) == 0 [similar to z-score, but divide by the square root of the sd]
            - 'minmax': min max normalisation i.e. min(Y[i,:] == 0, max(Y[i,:] == 1
            - 'mean': mean normalization i.e. mean(Y[i,:]) == 0, min(Y[i,:]) >= -1, max(Y[i,:]) <= 1
            - 'median': median normalisation i.e. median(Y[i,:]) == 0, min(Y[i,:]) >= -1, max(Y[i,:]) <= 1
            - 'sum': sum relative intensity (sum(Y[i,:]) == 1)
            - 'max': max relative intensity (max(Y[i,:]) == 1)
            - 'unit_vector': unit vector sum relative intensity (sum(Y[i,:].^2) == 1)
            - 'binary': create a presence/absence matrix of formula, ignoring peak intensity.
            - 'none': do not do normalise, return the original, or transform(original) input matrix.
    norm_transform: A string describing whether or not to transform the data prior to normalisation. One of:
            - 'none': do not transform the data 
            - 'log' : log transform the data 
            - 'power2' : square root power transform the data
            - 'power3' : cube root power transform the data
    norm_direction: A string describing which direction to normalise the data. One of:
            - 'rows': normalise the data row wise (this is done to account for differences between samples due to dilution effects)
            - 'columns': normalise the data column wise (this is done to account for fold differences in the concentration and variance of different metabolites)
    norm_subset: A string decribing the subset method to be used to generate the normalisation factors (tgnf). One of:
            - 'ALL': Use all formula in the dataset tgnf. If you are normalising column wise you must choose this option. 
            - 'LOS': take the top L order formula (ranked by intensity) from each sample tgnf, where p_L is the number of formula to take (default 500). 
            - 'PPP': take the proportion of peaks with a minimum percentage of observed values tgnf, where p_P defines the minimum percentage (default 0.5 (50%)). 
    p_L: parameter L in LOS subset, i.e. the number of top formula to take from each sample
    p_P: parameter P in PPP subset, i.e. the minimum percentage of observed intensities required to maintain a formula
        
    Info
    ----------
    For a description of unit vector sum relative intensity see:
    "On the Normalization of a Mass Spectrum for Comparison of Two Spectra - Zeev B. Alfassi ASMS (2004)"
    For a description of the other normalisation approaches used in this function, please refer to :
    "FT-ICR-MS Peak Intensity Normalization for Complex Mixture Analyses" Thompson et al. 2021.
    and "pmartR: Quality Control and Statistics for Mass Spectrometry-Based Biological Data", Stratton et al. 2019"
    For a good overview of the effect of centering, scaling and transformations on data see: 
    "van den Berg, Robert A., et al. "Centering, scaling, and transformations: improving the biological information content of metabolomics data. 2006."
    """
    #Tests
    ## CHECK THE INPUT METHODS
    assert norm_method in ['center', 'zscore','minmax','mean','median','sum','max','unit_vector','binary','none', 'pareto'], "norm method not recognised"
    assert norm_subset in ['ALL','LOS','PPP'], "subset method not recognised"
    assert norm_transform in ['none', 'log', 'power2', 'power3'], "transform method not recognised"
    assert norm_direction in ['rows', 'columns'], "norm direction method not recognised"
    if norm_direction == 'columns':
        assert norm_subset == 'ALL', "You can not subset the data if you are doing column wise normalisation"
    #Setup
    ## TRANSFORM THE INPUT DATA
    ordination_supplied = False
    if isinstance(input_matrix,pd.DataFrame): #if the user has supplied a dataframe e.g. produced by ordination_matrix
        ordination_supplied = True #later we will transform the data back to a dataframe
        ordination_copy = input_matrix.copy()
        input_matrix = input_matrix.to_numpy(dtype = float)
    oneDreshape = False
    if len(input_matrix.shape) == 1: #i.e. if the matrix is 1D
        input_matrix = np.reshape(input_matrix,(1,len(input_matrix))) #reshape to 2D 
        oneDreshape = True #later we will transform the data back to 1D
    if norm_transform == 'log': #log transform the data
        assert 0 not in input_matrix, "log 0 undefined, consider imputing 0 values as 1"
        input_matrix = np.log(input_matrix)
    elif norm_transform == 'power2':
        input_matrix = np.sqrt(input_matrix)
    elif norm_transform == 'power3':
        input_matrix = np.cbrt(input_matrix)
    ##CREATE THE TRANSFORMED MATRIX 
    if norm_direction == 'columns':
        input_matrix = input_matrix.T #transpose the matrix if we are normalising column wise
    rows,cols = np.shape(input_matrix)
    transformed_matrix = np.zeros((rows,cols))
    #Main
    ## POPULATE THE TRANSFORMED MATRIX
    if norm_method == 'none':
        transformed_matrix = input_matrix
    else:
        if norm_subset == 'PPP':
            # first calculate the columns above the PPP threshold
            threshold = int(p_P * rows)
            # create a booean array corresponding to the above p_P cols in each row
            boolean = np.array([False] * cols)
            for c in range(0,cols):
                if sum(input_matrix[:,c] > 0) > threshold:
                    boolean[c] = True
            assert sum(boolean) > 0, 'you have set the p_P value to high, there are no data points in your subset'
            print(f" There are {sum(boolean)} peaks/formula in your subset.")
        else:
            # create a boolean array of all columns 
            boolean = np.array([True] * cols)
        for i in range(0,rows):
            # create a boolean array corresponding to the top p_L values in this row
            if norm_subset == 'LOS':
                dsc_idx = np.argsort(input_matrix[i,:])[::-1] #sort the values in descending order
                dsc_idx = np.roll(dsc_idx,-np.count_nonzero(np.isnan(input_matrix[i,:]))) #put nans to the end of the array
                L_order = dsc_idx[0:p_L]
                boolean = np.array([False] * cols)
                for L in L_order:
                    boolean[L] = True #populate boolean array with top L values
            # calculate the appropriate normalisation factors                         
            row_sum = np.nansum(input_matrix[i,boolean])
            row_max = np.nanmax(input_matrix[i,boolean])
            row_min = np.nanmin(input_matrix[i,boolean])
            row_mean = np.nanmean(input_matrix[i,boolean])
            row_median = np.nanmedian(input_matrix[i,boolean])
            row_std = np.nanstd(input_matrix[i,boolean])
            row_norm = np.sqrt(np.nansum(input_matrix[i,boolean]**2))
            if norm_method == 'sum':
                for x in range(0,cols):
                        transformed_matrix[i,x] = input_matrix[i,x] / row_sum
            elif norm_method == 'max':
                for x in range(0,cols):
                        transformed_matrix[i,x] = input_matrix[i,x] / row_max
            elif norm_method == 'unit_vector':
                  for x in range(0,cols):
                        transformed_matrix[i,x] = input_matrix[i,x] / row_norm
            elif norm_method == 'zscore':
                  for x in range(0,cols):
                        transformed_matrix[i,x] = (input_matrix[i,x] - row_mean)/ row_std
            elif norm_method == 'pareto':
                  for x in range(0,cols):
                        transformed_matrix[i,x] = (input_matrix[i,x] - row_mean)/ np.sqrt(row_std)
            elif norm_method == 'minmax':
                  for x in range(0,cols):
                        transformed_matrix[i,x] = (input_matrix[i,x] - row_min)/ (row_max - row_min)
            elif norm_method == 'mean':
                  for x in range(0,cols):
                        transformed_matrix[i,x] = (input_matrix[i,x] - row_mean)/ (row_max - row_min)
            elif norm_method == 'median':
                  for x in range(0,cols):
                        transformed_matrix[i,x] = (input_matrix[i,x] - row_median)/ (row_max - row_min)
            elif norm_method == 'binary':
                    row_binary = input_matrix[i,:] > 0 
                    transformed_matrix[i,row_binary] = 1
            elif norm_method == 'center':
                  for x in range(0,cols):
                        transformed_matrix[i,x] = input_matrix[i,x] - row_mean
    ## RETRANSFORM THE DATA AND RETURN
    if norm_direction == 'columns':
        transformed_matrix = transformed_matrix.T
    if oneDreshape == True: #if input_matrix provided as 1D
        transformed_matrix = np.reshape(transformed_matrix,(transformed_matrix.shape[1],)) #reshape back to 1D
    if ordination_supplied == True: #if input_matrix supplied as pk.ordination_matrix
        ordination_copy.iloc[:,:] = transformed_matrix #repopulate the ordination matrix with transformed values
        return ordination_copy #return an ordination matrix
    else:
        return transformed_matrix #else return a numpy array



from numpy.lib.function_base import _rot90_dispatcher, rot90
from ..diversity.normalise_intensity import normalise_intensity
from ..formula.calculate_mass import calculate_mass
from matplotlib import pyplot as plt
import pandas as pd
import numpy as np

def mass_spectrum(msTuple, 
                  method = 'monoisotopic',
                  logTransform = False,
                  stepSize = 5,
                  lineColor = 'g',
                  lineWidth = '0.8',
                  invertedAxis = [],
                  invertedAxisLabel = 'Inverted Axis Label',
                  invertedAxisColor = 'b',
                  invertedAxisLineWidth = 0.8,
                  **kwargs):
    """ 
    Docstring for function PyKrev.mass_spectrum
    ====================
    This function takes an msTuple and plots a mass spectrum using atomic masses calculated via method.  
    
    Use
    ----
    mass_spectrum(Y,peak_intensities)
    
    Returns the figure and axes handles
    
    Parameters
    ----------
    Y: msTuple
    method: the method used to calculate formula mass. See pk.calculate_mass for more information. 
            one of
            'monoisotopic'
            'average'
            'nominal'
            or
            'mz'- plot a list of mz values in msTuple
    logTransform: boolean, log transform the y axis
    stepSize: int, number of decimal places to round masses to 
    lineColor: string, color for primary axis
    lineWidth: int, line width for primary axis
    invertedAxis: np.ndarray, array of values to plot on the negative y axis, should all be positive
    invertedAxisLabel: string, label for negative y axis
    invertedAxisColor: string, color for inverted axis
    invertedAxisLineWidth: int, line width for inverted axis
    **kwargs: key word arguments to plt.plot, must not include color or linewidth arguments.
    """
    #Tests
    ## Set default values for color and linewidth unless they have been given in kwargs
    assert 'color' not in kwargs, 'provide colo(u)r as lineColour'
    assert 'lineWidth' not in kwargs, 'provide linewidth as lineWidth'
    assert method in ['monoisotopic','average','nominal','mz'], 'Provide a valid method. See docstring for info.'
    #Setup
    peak_intensities = msTuple.intensity
    mz_list = msTuple.mz
    if logTransform:
        peak_intensities = np.log(peak_intensities)
    if method == 'mz':
        mass = mz_list
    else: 
        mass = calculate_mass(msTuple, method = method)
    if len(invertedAxis) > 0:
        assert len(invertedAxis) == len(peak_intensities), 'inverted data must be the same length as peak intensity array'
        peak_intensities = [x for _,x,_ in sorted(zip(mass,peak_intensities, invertedAxis))]
        mass = [y for y,_,_ in sorted(zip(mass,peak_intensities, invertedAxis))]
        invertedAxis = [z for _,_,z in sorted(zip(mass,peak_intensities, invertedAxis))]
    else:
        peak_intensities = [x for _,x in sorted(zip(mass,peak_intensities))]
        mass = [y for y,_ in sorted(zip(mass,peak_intensities))]
    assert len(peak_intensities) == len(mass)
    ## create plotting mass and plotting peak intensity variables 
    ### step size should be based on the precision of the masses
    ### we can calculate an appropriate step size by finding the minimum difference between adjacent masses
    ### massDiff = np.append(mass,0)[1::] - np.array(mass)
    ### stepSize = str(min(massDiff[::-1]))[::-1].find('.')
    plotting_mass = np.arange(start=min(mass) - 5,stop = max(mass) + 5, step = 10 ** -stepSize)
    plotting_intensity = np.zeros(len(plotting_mass))
    indices = np.nonzero(np.isin(np.around(plotting_mass,stepSize),np.around(mass,stepSize)))[0]
    assert len(indices) == len(mass), f"{len(indices)} {len(mass)} increase stepSize"
    if len(invertedAxis) > 0:
        plotting_inverted = np.zeros(len(plotting_mass))
        for idx,iy,inv in zip(indices,peak_intensities, invertedAxis):
            plotting_intensity[idx] = iy
            # this approach only works with positive values
            if inv < 0:
                inv = 0
            plotting_inverted[idx] = inv
    else:
        for idx,iy in zip(indices,peak_intensities):
            plotting_intensity[idx] = iy
    #Main
    fig, ax1 = plt.subplots()
    ax1.plot(plotting_mass,plotting_intensity,color=lineColor, linewidth=lineWidth, **kwargs)
    if len(invertedAxis) > 0: 
        ax2 = ax1.twinx()
        ax2.plot(plotting_mass,plotting_inverted,color=invertedAxisColor, linewidth=invertedAxisLineWidth, **kwargs)
        ax1.set_ylabel("Intensity", loc = "top")
        ax2.set_ylabel(f"{invertedAxisLabel}", loc = "bottom")
        ax2.grid(axis='y', alpha=0.75)
        ## The following is a bit hacky but I couldn't find a better way. Essentially mirror the axes, and remove the negative tick marks.
        ### First remove negative ticks 
        ax1ticks = [tick for tick in ax1.get_yticks() if tick >=0]
        ax2ticks = [tick for tick in ax2.get_yticks() if tick >=0]
        ### Add a mirrored max negative value
        ax1.set_ylim(ymin=-max(ax1ticks), ymax=max(ax1ticks))
        ax2.set_ylim(ymin=-max(ax2ticks), ymax=max(ax2ticks))
        ### Set tick values
        ax1.set_yticks(ax1ticks)
        ax2.set_yticks(ax2ticks)
        ### Invert ax2
        ax2.invert_yaxis()
        ### Remove the tick label and tick marks for this value
        #yticks = ax1.yaxis.get_major_ticks()
        #yticks[-1].label1.set_visible(False)
        #yticks[-2].label1.set_visible(False)
        #yticks[-1].tick1line.set_visible(False)
        #yticks[-2].tick1line.set_visible(False)
        #yticks = ax2.yaxis.get_major_ticks()
        #yticks[-1].label1.set_visible(False)
        #yticks[-2].label1.set_visible(False)
        #yticks[-1].tick1line.set_visible(False)
        #yticks[-2].tick1line.set_visible(False)
    else:
        ax1.set_ylabel("Intensity")
    ax1.grid(axis='y', alpha=0.75)
    if method == 'mz':
        ax1.set_xlabel('m/z')
    else:
        ax1.set_xlabel(f"{method[0].upper()}{method[1::]} atomic mass")
    if len(invertedAxis) > 0: 
        return fig, ax1, ax2
    else:
        return fig,ax1



import pandas as pd
import numpy as np
from ..diversity.ordination_matrix import ordination_matrix
from .find_intersections import find_intersections
from .average_mstuple import average_mstuple

class msTupleDict(dict):
    """ 
    Docstring for class pykrev.msTupleDict
    ==========
    Core DataType for PyKrev batch processing. 
    A dictionary with sample names as keys and msTuple objects as values.

    Use
    ----------
    msTupleDict() 
    
    Returns an empty msTupleDict. 
    
    Methods 
    ----------
    msTupleDict.validate(): validate all of the msTuples stored in the dictionary.

    msTupleDict.summary(): summarise all of the msTuples stored in the dictionary.

    msTupleDict.subset(subsetList = []): subset the dictionary based on the keys found in subsetList

    msTupleDict.average(intensityMethod = 'mean', mzMethod = 'mean', minOccurrence = 1, zeroValues = True, stdDev = False): return an average msTuple 

    msTupleDict.intersections(exclusive = True): return a dictionary contanining all intersections between the formula in msTupleDict. See pk.find_intersections.

    msTupleDict.to_OrdinationMatrix(impute_value = 'nan'): write the contents of the msTupleDict to an ordination matrix. See pk.ordination_matrix. 

    msTupleDict.to_DataFrame(): write the contents of the msTupleDict to a pandas dataframe. Columns are 'assigned formula', 'mean mz' and 'std mz'

    """
    def validate(self):
        for v in self.values():
            v.validate()

    def summary(self):
        print(f'msTupleDict containing {len(self)} samples.')
        print()
        for k,v in zip(self.keys(),self.values()):
            print(f'{k} summary')
            print(f'{"*"*20}')
            v.summary()
            print()

    def subset(self, subsetList = []):
        self.validate()
        return msTupleDict({key: value for key, value in self.items() if key in subsetList})

    def average(self, intensityMethod = 'mean', mzMethod = 'mean', minOccurrence = 1, zeroValues = True, stdDev = False):
        self.validate()
        return average_mstuple(self, intensityMethod = intensityMethod, mzMethod = mzMethod, minOccurrence = minOccurrence, zeroValues = zeroValues, stdDev = stdDev)

    def intersections(self, exclusive = True):
        self.validate()
        return find_intersections(self,exclusive = exclusive)
    
    def to_DataFrame(self):
        self.validate()
        df = pd.DataFrame(index=self.keys())
        for k,v in zip(self.keys(),self.values()):
            df.loc[k,'assigned formula'] = len(v.formula)
            df.loc[k,'mean mz'] = np.mean(v.mz)
            df.loc[k,'std mz'] = np.std(v.mz)
        return df
    
    def to_OrdinationMatrix(self, impute_value = 'nan'):
        self.validate()
        return ordination_matrix(self, impute_value = impute_value)




import pandas as pd
import numpy as np 
from .msTuple import msTuple
def read_corems(corems_df, mass_type = 'calibrated'):
    """ 
    Docstring for function PyKrev.read_corems
    ====================
    This function reads a pandas dataframe produced in corems using mass_spectrum_obj.to_dataframe() and outputs an msTuple.
    
    Use
    ----
    read_corems(corems_df)
    
    Returns an msTuple
    
    Parameters
    ----------
    mass_type: String, What form to read the mass from the dataframe in. One of: 
        'calculated' (i.e. exact formula mass),
        'calibrated' (experimental mass after calibration)
        'experimental' (raw mass from spectrum).

    Info
    -----------    
    Corems returns isotope assignments that pykrev cannot parse. 
    PyKrev will filter out formula with isotopes assigned.
    """
    #Tests
    assert(mass_type in ['calculated','calibrated','experimental']), 'incorrect mass_type given'
    #Setup
    assigned = corems_df['Is Isotopologue'] == 0
    assignedDf = corems_df[assigned].copy()
    N = assignedDf.shape[0]
    #Main
    ## first check that there are columns in the dataframe corresponding to C,H,N,O,P,S if there aren't, we make them!
    if 'C' not in assignedDf.columns:
        assignedDf['C'] = [""] * N
    if 'H' not in assignedDf.columns:
        assignedDf['H'] = [""] * N
    if 'N' not in assignedDf.columns:
        assignedDf['N'] = [""] * N      
    if 'O' not in assignedDf.columns:
        assignedDf['O'] = [""] * N           
    if 'S' not in assignedDf.columns:
        assignedDf['S'] = [""] * N       
    if 'P' not in assignedDf.columns:
        assignedDf['P'] = [""] * N      
    Cstring = []
    for cnumber in assignedDf['C']:
        try: 
            if int(cnumber) == 0:
                Cstring.append("")
            else:
                Cstring.append(str(f"C{int(cnumber)}"))
        except ValueError:
            Cstring.append("")
    Hstring = [] 
    for hnumber in assignedDf['H']:
        try: 
            if int(hnumber) == 0:
                Hstring.append("")
            else:
                Hstring.append(str(f"H{int(hnumber)}"))
        except ValueError:
            Hstring.append("")        
    Nstring = [] 
    for nnumber in assignedDf['N']:
        try:
            if int(nnumber) == 0:
                Nstring.append("")
            else:
                Nstring.append(str(f"N{int(nnumber)}"))
        except ValueError:
            Nstring.append("")
    Ostring = [] 
    for onumber in assignedDf['O']:
        try:
            if int(onumber) == 0:
                Ostring.append("")
            else:
                Ostring.append(str(f"O{int(onumber)}"))
        except ValueError:
            Ostring.append("")
    Pstring = []
    for pnumber in assignedDf['P']:
        try:
            if int(pnumber) == 0:
                Pstring.append("")
            else: 
                Pstring.append(str(f"P{int(pnumber)}"))
        except ValueError:
            Pstring.append("") 
    Sstring = [] 
    for snumber in assignedDf['S']:
        try:
            if int(snumber) == 0:
                Sstring.append("")
            else: 
                Sstring.append(str(f"S{int(snumber)}"))
        except ValueError:
            Sstring.append("")
    pykrev_formula = [c + h + n + o + p + s for c, h, n, o, p, s in zip(Cstring, Hstring, Nstring, Ostring, Pstring, Sstring)]
    pykrev_abundance = assignedDf['Peak Height']
    if mass_type == 'calibrated':
        pykrev_mass = assignedDf['Calibrated m/z']
    elif mass_type == 'calculated':
        pykrev_mass = assignedDf['Calculated m/z']
    else:
        pykrev_mass = assignedDf['m/z']
    return msTuple(pykrev_formula, np.array(pykrev_abundance), np.array(pykrev_mass))


import pandas as pd
import numpy as np
def ordination_matrix(msTupleDict, impute_value = 'nan'):
    """ 
	Docstring for function pyKrev.ordination_matrix
	====================
	This function computes a sample data matrix from an msTupleDict
    This matrix can be used for further ordination analysis (e.g. PCA, PCoA...)
    
	Use
	----
	ordination_matrix(Y)
    
	Returns a pandas dataframe in which the column headers are a set of all formula found in msTupleDict and the rows correspond to a specific sample.
    The [row,col] value of the dataframe is therefore the peak intensity of a formula. Impute value (default 0) if the formula was not present. 
    
	Parameters
	----------
	Y: an msTupleDict
    impute_value: the value to impute when a formula isn't present in a group. An integer or float or 'nan' (default 0):
    """
    #Setup
    if  impute_value == 'nan':
        impute_value = np.nan
    group_names = msTupleDict.keys()
    molecular_formulas = []
    peak_intensities = []
    for msTuple in msTupleDict.values():
        molecular_formulas.append(msTuple[0])
        peak_intensities.append(msTuple[1])
    all_formula = set()
    for flist in molecular_formulas:
        all_formula.update(flist) #update the set with the formula in molecular_formula
    all_formula = list(all_formula) #convert back to list for ease with indexing 
    ordination_mat = pd.DataFrame(columns = all_formula,index = group_names)
    col_index = 0
    #Main
    for f in all_formula: #cycle through all_formula 
        row_index = 0 
        for flist, plist in zip(molecular_formulas,peak_intensities):
            try:
                i = flist.index(f) #find the first instance of the formula in flist
                ordination_mat.iloc[row_index,col_index] = plist[i] #set the row value to the peak intensity for that sample 
                row_index += 1  #move down a row i.e. sample
            except ValueError: #if the formula can't be found 
                ordination_mat.iloc[row_index,col_index] = impute_value #set the value to impute_value 
                row_index += 1 #move down a row
        col_index += 1 #move across to the next column i.e. formula
    return ordination_mat



import itertools
import numpy as np
import pandas as pd
def find_intersections(msTupleDict, exclusive = True):
    """   
	Docstring for function pykrev.find_intersections
    
	====================
	This function compares n msTuples inside an msTupleDict and outputs a dictionary containing the intersections between each msTuple.
    
	Use
	----
	find_intersections(Y)
    
	Returns a dictionary in which each key corresponds to a combination of sample names 
	and the corresponding value is a set containing the intersections between the groups in that combination.  
    
	Parameters
	----------
	Y: msTupleDict
	exclusive: Boolean, True or False, depending on whether you want the intersections to contain only unique values.
    """
    #Setup
    formula_lists = []
    group_labels = msTupleDict.keys()
    for msTuple in msTupleDict.values():
        formula_lists.append(msTuple.formula)
    combinations = [seq for i in range(0,len(group_labels)+1) for seq in itertools.combinations(group_labels,i) if len(seq) > 0]
    combinations = sorted(combinations,key = lambda c : len(c),reverse = True) # sort combinations by length
    if exclusive == True:
        assigned_formula = set() #create a set that will hold all the formula already assigned to a group
    amb = pd.DataFrame(data = formula_lists).T
    amb.columns = group_labels
    intersections = dict() 
    #Main
    for combo in combinations:         
            queries = [] 
            for c in combo:
                formula = list(filter(None,amb[c])) #Remove None entries introduced by dataframe
                queries.append(set(formula)) 
            if len(queries) == 1: #if there is only one query find the unique elements in it 
                q_set = frozenset(queries[0]) #qset is a frozen set, so it will not be mutated by changes to queries[0]
                for f_list in formula_lists: #cycle all formula in formula_lists
                    set_f = frozenset(f_list) #convert f_list to sets, must be frozen so type matches q_set
                    if set_f == q_set: # ignore the set that corresponds to the query 
                        pass
                    else:
                        queries[0] = queries[0] - set_f #delete any repeated elements in fset
                            
                intersections[combo] = queries[0] 
                
            elif len(queries) > 1: 
                if exclusive == True:
                    q_intersect = intersect(queries)
                    intersections[combo] = q_intersect - assigned_formula #remove any elements from q_intersect that have already been assigned
                    assigned_formula.update(q_intersect) #update the assigned_set with q_intersect
                else:
                    intersections[combo] = intersect(queries)  
    return intersections
    
def intersect(samples,counter=0):
    """ This command uses recursion to find the intersections between a variable number of sets given in samples. 
        Where samples = [set_1,set_2,...,set_n] """
    if len(samples) == 1: 
        return samples[0]
    a = samples[counter]
    b = samples[counter+1::]
    if len(b) == 1: #check to see whether the recursion has reached the final element
        return a & b[0]
    else:
        counter += 1 
        return a & intersect(samples,counter)


from typing import NamedTuple
import numpy as np
import pandas as pd
class msTuple(NamedTuple):
    """ 
    Docstring for class pykrev.msTuple
    ==========
    Core DataType for PyKrev. A tuple representing an assigned mass spectrum and contains three objects:
            1. msTuple.formula: a list of molecular formula strings 
            2. msTuple.intensity: a numpy.ndarray of peak intensities
            3. msTuple.mz a numpy.nd array of calibrated mz values
    
    Use
    ----------
    msTuple(Y,Z,X) 
    
    Returns an msTuple
    
    Parameters 
    ----------
    Y: A list of molecular formula strings. 
    Z: A numpy.ndarray of peak intensities
    X: A numpy.ndarray of mz values.  
    
    Methods 
    ----------
    msTuple.validate(): validate the type and length of objects in an msTuple
    
    msTuple.summary(): summarise the objects in the tuple

    msTuple.filter_mz(low,high): returns a new msTuple filtered between low and high mz

    msTuple.filter_intensity(low,high): returns a new msTuple filtered between low and high intensity

    msTuple.filter_spectral_interference(): returns a new msTuple which has been filtered of spectral interference by doubley charged molecular ions (see pykrev.filter_spectral_interference)

    msTuple.to_csv(): writes the msTuple to a .csv file
    """
    
    formula: list
    intensity: np.ndarray
    mz: np.ndarray
    
    def __repr__(self) -> str:
        return f'msTuple(formula={np.array(self.formula)}), intensity={self.intensity}, mz={self.mz}'
    
    def validate(self):
        assert type(self.formula) == list, "msTuple.formula must be provided as a list"
        assert type(self.intensity) == np.ndarray, "msTuple.intensity must be provided as an ndarray"
        assert type(self.mz) == np.ndarray, "msTuple.mz must be provided as an ndarray"
        assert len(self.formula) > 1, "msTuple.formula must not be empty"
        assert len(self.formula) == len(self.intensity) == len(self.mz), "msTuple.formula, msTuple.intensity and msTuple.mz must be of equal length"
    
    def summary(self):
        self.validate()
        print(f'assigned formula = {len(self.formula)} ')
        print(f'min intensity = {np.min(self.intensity):.1E}')
        print(f'max intensity = {np.max(self.intensity):.1E}')
        print(f'mean mz = {np.mean(self.mz)} ')
        print(f'std mz = {np.std(self.mz):.1E}')
    
    def filter_spectral_interference(self, tol = 2, verbose = True):
        #Setup
        self.validate()
        mass_list = self.mz
        formula_list = np.array(self.formula)
        peak_intensities = self.intensity
        mass_defect = mass_list - np.floor(mass_list)
        c13d2 = (13.00335-12)/2
        flags = np.zeros(len(mass_list))
        #Main
        for i in range(0,len(mass_list)):
            if mass_defect[i] > 0.4 and mass_defect[i] < 0.8:
                flags[i] = 1
                mono_mass = mass_list[i] - c13d2
                mass_diff = abs(mass_list - mono_mass)
                mono_pos = mass_diff == min(mass_diff)
                if min(mass_diff)/mass_list[i] * 1e6 < tol:
                    if peak_intensities[mono_pos]/peak_intensities[i] < 10:
                        flags[mono_pos] = 2
        spectralFilter = flags == 0 
        filtermz = mass_list[spectralFilter]
        filterformula = formula_list[spectralFilter]
        filterintensity = peak_intensities[spectralFilter]
        if verbose == True:
            print(f"{len(mass_list)-len(filtermz)} interferences removed.")
        filterformula = list(filterformula) #reconvert filter_formula to list to sort out data type issue
        return self._replace(formula = filterformula, intensity = filterintensity, mz = filtermz)
    
    def filter_mz(self,low,high):
        self.validate()
        highpassBool = self.mz > low
        lowpassBool = self.mz < high
        bandpassBool = highpassBool & lowpassBool
        fArray = np.array(self.formula)
        filtermz = self.mz[bandpassBool]
        filterintensity = self.intensity[bandpassBool]
        filterformula = list(fArray[bandpassBool])
        return self._replace(formula = filterformula, intensity = filterintensity, mz = filtermz)
    
    def filter_intensity(self,low,high):
        self.validate()
        highpassBool = self.intensity > low
        lowpassBool = self.intensity < high
        bandpassBool = highpassBool & lowpassBool
        fArray = np.array(self.formula)
        filtermz = self.mz[bandpassBool]
        filterintensity = self.intensity[bandpassBool]
        filterformula = list(fArray[bandpassBool])
        return self._replace(formula = filterformula, intensity = filterintensity, mz = filtermz)

    def to_csv(self, path):
        self.validate()
        csv = pd.DataFrame()
        csv['formula'] = self.formula
        csv['intensity'] = self.intensity
        csv['m/z'] = self.mz
        csv.to_csv(path, index= False)
        print(f'msTuple written to {path}')


from .calculate_mass import calculate_mass
from .element_counts import element_counts
import numpy as np
def kendrick_mass_defect(msTuple, base = 'CH2', rounding = 'even'):
    """ 
	Docstring for function pyKrev.kendrick_mass_defect
	====================
	This function takes an msTuple and calculates kendrick mass defect scores.  
    
	Use
	----
	kendrick_mass_defect(Y)
    
	Returns a tuple contains two numpy arrays of len(Y[2]). The first contains the kendrickMass and the second contains the kendrickMassDefect.
    
	Parameters
	----------
    Y: msTuple 
    
    base: String, atom group used to define the Kendrick mass.
    
    rounding: The method of rounding to use when calculating kendrick mass defect. One of 
        - even: round to the nearest integer value. for values exactly halfway between rounded decimal values, numpy rounds to the nearest even value. 
                Thus 1.5 and 2.5 round to 2.0, -0.5 and 0.5 round to 0.0, etc.
        - rint: round to the nearest integer value. for values exactly halfway between rounded decimal values, numpy rounds up to the nearest integer value. 
                Thus 1.4 becomes 1.0, 1.5 becomes 2.0, 2.5 becomes 3.0 
        - floor: always round down to the nearest integer value. Thus 1.4 becomes 1.0, 2.6 becomes 2.0.
        - ceil: always round up to the nearest integer value. Thus 1.4 becomes 2.0, 2.4 becomes 3.0. 
    Info
	----------
    Calculation taken from Hughey et al (2001) 
    "Kendrick Mass Defect Spectrum: A Compact Visual Analysis for Ultrahigh-Resolution Broadband Mass Spectra"
    Rounding calculations may lead to "edge effects" in complex datasets with many peaks. 
    We recommend experimenting with different rounding methods when making kmd plots.
    """
    #Tests
    assert(rounding in ['even','rint','floor','ceil']), "provide a valid rounding method"
    #Setup
    if type(base) == str:
        base = [base]
    mz_list = msTuple[2]
    kendrickMass = np.empty(len(mz_list))
    kendrickMassDefect = np.empty(len(mz_list))
    #Main
    nominalBase = calculate_mass(base, method = 'nominal')
    exactBase = calculate_mass(base, method = 'monoisotopic')
    kendrickMass = np.array(mz_list) * (nominalBase/exactBase)
    if rounding == 'even':
        kendrickMassDefect  = np.around(kendrickMass) - kendrickMass
    elif rounding == 'rint':
        kendrickMassDefect  = np.rint(kendrickMass) - kendrickMass
    elif rounding == 'floor':
        kendrickMassDefect  = np.floor(kendrickMass) - kendrickMass
    elif rounding == 'ceil':
        kendrickMassDefect  = np.ceil(kendrickMass) - kendrickMass
    return kendrickMass, kendrickMassDefect


import numpy as np
from .element_counts import element_counts
def element_ratios(msTuple, ratios = ['OC','HC'], zero_ratio = True):
        
    """ 
	Docstring for function pyKrev.element_ratios
	====================
	This function takes an msTuple and returns atomic ratios of each ratio given in ratios for each formula in the formula list.
    
	Use
	----
	element_ratios(Y)
    
	Returns a list of len(Y) in which each item is a dictionary containing the atomic ratios listed in ratios, 
	calculated as the first element / second element. If the second element is zero, a nan or zero value is returned
    depending on zero_ratio.
    
	Parameters
	----------
	Y: msTuple OR a list of molecular formula strings
    
	ratios: List, atomic ratios to compute, calculated as first element divided by second element. Only C,H,N,O,P & S can be used. 
    
    zero_ratio: Boolean, behaviour for if a zero ratio is encountered (i.e. the denominator is zero), if true a zero value is returned. 
    
    """
    count_list = element_counts(msTuple)
    ratio_list = []
    for count in count_list:
        ratio_counts = dict()
        for atom in ratios:
            if len(atom) > 2:
                raise Exception('Only C,H,N,O,P & S can be used. ')
            if count[atom[1]] == 0:
                if zero_ratio == True:
                    ratio_counts[atom] = 0
                else:
                    ratio_counts[atom] = np.nan
            else:
                ratio_counts[atom] = count[atom[0]]/count[atom[1]]
        ratio_list.append(ratio_counts)    
    return ratio_list


import pandas as pd
import numpy as np 
from .msTuple import msTuple
from .msTupleDict import msTupleDict
from ..diversity.ordination_matrix import ordination_matrix
def read_batch_formularity(report_name):
    """ 
	Docstring for function PyKrev.read_batch_formularity
	====================
	This function reads in data from the report csv file produced by formularity software when performing batch assignment with alignment.
	The function filters out any formula that do not have C and H atoms.
    
	Use
	----
	read_batch_formularity(report_name)
    
	Returns an msTupleDict contaning sample names as keys and corresponding msTuples as values 

    
	Parameters
	----------
	report_name: name of csv file that the formularity report to be read is saved as. 
    
    Info
    -----------
    PyKrev will filter out formula with 13C assignments
    """
    #Main
    report = pd.read_csv(report_name)
    notIsotopologue = report['C13'] == 0 # boolean array of only non isotopologues
    report = report[notIsotopologue]
    report.reset_index(drop = True, inplace = True) #reset the index to account for the removed isotopologues]
    C = report['C']
    H = report['H']
    O = report['O']
    N = report['N'] 
    S = report['S']
    P = report['P']
    #dimensions of report (number of formula, samples)
    x,y = report.shape
    #list to store all formula in
    molecular_formula = []
    #turn atom numbers into formula strings 
    for n in range(0,len(C)):
        if C[n] == 0 and H[n] == 0: #If there isn't a count for C and H don't include the formula
            temp_formula = np.nan
        else:
            temp_formula = 'C'+str(C[n])+'H'+str(H[n]) #add C and H then ... 
            if N[n] > 0:
                temp_formula = temp_formula + 'N' + str(N[n])
            if O[n] > 0:
                temp_formula = temp_formula + 'O' + str(O[n])
            if P[n] > 0:
                temp_formula = temp_formula + 'P' + str(P[n])
            if S[n] > 0:
                temp_formula = temp_formula + 'S' + str(S[n])
        molecular_formula.append(temp_formula)
    #convert to numpy array so we can use boolean indexing
    molecular_formula = np.array(molecular_formula)
    #remove nan values
    filter_formula = molecular_formula[molecular_formula != 'nan']
    #add to the original dataframe to preserve indexing
    report['all_formula'] = molecular_formula
    #make new dataframe
    formulaData = pd.DataFrame(index=None)
    #copy across formula, mass, cclass and mass error data with boolean indexing
    formulaData['formula'] = report['all_formula'][molecular_formula != 'nan']
    formulaData['mass'] = report['Mass'][molecular_formula != 'nan']
    formulaData['cclass'] = report['Class'][molecular_formula != 'nan']
    formulaData['massError'] = report['Error_ppm'][molecular_formula != 'nan']
    spectraNames = []
    spectraFormula = []
    spectraIntensities = []
    #column 15 is the first sample ID column
    for i in range(14,y):
        #save the peak intensities after boolean indexing has been applied
        intensity = report.iloc[:,i][molecular_formula != 'nan']
        formulaData[report.columns[i]] = intensity
        spectraNames.append(report.columns[i])
        spectraIntensities.append(np.array(intensity[intensity > 0]))
        spectraFormula.append(list(filter_formula[intensity > 0]))
    #Now process the formulaData into an msTuple dictionary 
    batchDict = msTupleDict()
    x,y = formulaData.shape
    sampleNames = formulaData.columns[4::]
    for name in sampleNames:
        batchDict[name] = msTuple(list(formulaData['formula'][formulaData[name] > 0]), np.array(formulaData[name][formulaData[name] > 0]), np.array(formulaData['mass'][formulaData[name] > 0]))
    return batchDict



from ..formula.calculate_mass import calculate_mass
import numpy.linalg as la
import numpy as np
def page_rank(msTuple, reactionDict = {
                                            'decarboxylation': -calculate_mass(['CO2']),
                                            'methylation': calculate_mass(['CH2']),
                                            'demethylation': -calculate_mass(['CH2']),
                                            'hydrogenation': calculate_mass(['H2']),
                                            'dehydrogenation': -calculate_mass(['H2']),
                                            'hydration': calculate_mass(['H2O']),
                                            'dehydration': -calculate_mass(['H2O']),
                                            'oxidation': calculate_mass(['O']),
                                            'reduction': -calculate_mass(['O'])
                                            }, reactionWeights = {}, d = 0.9, tol = 0.01, roundVal = 8):
    """ 
	Docstring for function PyKrev.page_rank
	====================
	This function takes an msTuple and performs the pagerank algorithm on a reaction network derived from  the list of formula. 
    The reaction network can have different weights for each reaction type this should be provided in the dictionary reaction weights.
    For more information on the page rank algorithm and it's implementation here, see the docs: pykrev/docs/pagerank_and_networkvis/PageRankandNetworkVis.ipynb
    
	Use
	----
	page_rank(Y)
    
	Returns a numpy array of len(Y[0]) with the pagerank scores corresponding to the elements in Y[0].
    
	Parameters
	----------
	Y: msTuple
    reactionDict: dictionary, containing reaction names as keys and their associated change in monoisotopic formula mass as values.
    reactionWeights: dictionary, containing the relative weighting to give to each reactionType. If not provided each reactionWeight is given with equal value.
    d: float, damping factor in page rank algorithm
    tol: float, tolerance to run power iteration method to
    roundVal: int, number of digits to round to gor mass defect calculations
    """ 
    #Tests
    if len(reactionWeights) == 0:
        for key in reactionDict.keys():
            reactionWeights[key] = 1
    else: 
        assert reactionWeights.keys() == reactionDict.keys(), "reactionWeights and reactionKeys must have identical keys"
    #Setup
    formulaList = msTuple[0]
    N = len(formulaList)
    formulaMass = calculate_mass(formulaList) # Compute the exact monoisotopic mass of each formula in the dataset
    reactionList = list(reactionDict.keys())
    #Main
    ## Create the matrix L 
    L = np.zeros([N,N])
    for j in range(N):
        for reactionType in reactionList:
            # find possible matches, use np.round to account for rounding errors
            matchIndex = np.where(np.round(formulaMass,roundVal) == np.round((formulaMass[j] + reactionDict[reactionType]),roundVal))
            # if a match is found
            if len(matchIndex[0]) > 0:
                L[matchIndex,j] = reactionWeights[reactionType]
        ## normalise the probabilities so they sum to one
        if sum(L[:,j]) > 0:
            L[:,j] = L[:,j]/sum(L[:,j])
        ## if the column sums to zero set to 1/N
        else: 
            L[:,j] = 1/N
    # run the pageRank algorithm to convergence
    r = 100 * np.ones(N) / N # Sets up the probability vector
    M = d * L + (1-d)/N * np.ones([N, N]) 
    lastR = r
    r = M @ r
    i = 0
    while la.norm(lastR - r) > tol :
        lastR = r
        r = M @ r
        i += 1
    print(str(i) + " iterations to convergence.")
    return r


import numpy as np
from scipy import spatial
def bray_curtis_matrix(matrix):
    """ 
	Docstring for function pyKrev.bray_curtis_matrix 
	====================
	This function Scipy bray curtis function to compute a dissimilarity matrix of dimensions row * row.  
    
	Use
	----
	bray_curtis_matrix(Y)
    
	Returns a numpy array of shape(len(Y[:,0]),len(Y[:,0])) in which each value is the bray curtis dissimilarity value.  
    
	Parameters
	----------
	Y: A numpy array containing peak intensities - where rows correspond to samples and columns correspond to molecular formula
    
	Info
	----------
	The Bray-Curtis dissimilarity is always a number between 0 and 1. If 0, the two samples share all the same formula; if 1, they don’t share any formula. 
        
    """  
    assert(isinstance(matrix,np.ndarray)), 'must provide a numpy array'
    assert(len(matrix.shape) != 1), 'must provide at least two columns'
    row,col = np.shape(matrix)
    transformed_matrix = np.zeros((row,row))
    for x in range(0,row):
        for i in range(0,row):
            transformed_matrix[i,x] = spatial.distance.braycurtis(matrix[x,:],matrix[i,:])
    return transformed_matrix


import numpy as np
from .element_counts import element_counts
def nominal_oxidation_state(msTuple):
    """ 
	Docstring for function pyKrev.nominal_oxidation_state
	====================
	This function takes an msTuple and returns the nominal oxidatate state of C.
    
	Use
	----
	nominal_oxidation_state(Y)
    
	Returns a numpy array of len(Y[0]) in which each item, i, is the nominal oxidation state of C corresponding to Y[0][i].  
    
	Parameters
	----------
	Y: msTuple
    
	Info
	----------
	This function derives the nominal oxidation state of C as described in LaRowe, Douglas E., and Philippe Van Cappellen. 
	"NOSC allows for computation of the average carbon oxidation state of an organic compound without writing out the oxidation half reaction."
	"Degradation of natural organic matter: a thermodynamic analysis." 
	Geochimica et Cosmochimica Acta 75.8 (2011): 2030-2042.       

	NOSC = -((-Z + 4a + b - 3c - 2d + 5e - 2f)/a) + 4 
	where Z = charge and 
	a = C
	b = H
	c = N 
	d = O
	e = P
 	f = S
        
    """      
	#Setup
    count_list = element_counts(msTuple)
    NOSCs = np.array([])
    #Main
    for i in count_list:
        NOSCs = np.append(NOSCs,-((4*i['C'] + i['H'] - 3 * i['N'] - 2 * i['O'] + 5 * i['P'] - 2 * i['S'])/i['C']) + 4) 
    return NOSCs


from .diversity_indices import diversity_indices
from .normalise_intensity import normalise_intensity
from .bray_curtis_matrix import bray_curtis_matrix
from .ordination_matrix import ordination_matrix
from .compound_class import compound_class
from .page_rank import page_rank


import numpy as np
from .msTuple import msTuple
def filter_spectral_interference(msTupleObj, tol = 2 , verbose = True):
    """ 
	Docstring for function pykrev.filter_spectral_interference
	====================
	This function takes an msTuple and returns filtered copies of all arrays with suspected spectral interferences due to high molecular weight doubly charged ions, removed.
    
	Use
	----
	filter_spectral_interference(Y)
    
	Returns an msTuple containing filtered arrays.
    
	Parameters
	----------
    Y: msTuple
    tol: Integer of float,the tolerance (in ppm) used to identify monoisotopic 
    verbose: Booloean, print text to the console.

    Info
    -----------
    For more information on this algorithm refer to Patriarca, Claudia, and Jeffrey A. Hawkes. 
    "High Molecular Weight Spectral Interferences in Mass Spectra of Dissolved Organic Matter." 
    Journal of the American Society for Mass Spectrometry (2020).
    """    
    #Setup
    mass_list = msTupleObj[2]
    formula_list = np.array(msTupleObj[0])
    peak_intensities = msTupleObj[1]
    mass_defect = mass_list - np.floor(mass_list)
    c13d2 = (13.00335-12)/2
    flags = np.zeros(len(mass_list))
    #Main
    for i in range(0,len(mass_list)):
        if mass_defect[i] > 0.4 and mass_defect[i] < 0.8:
            flags[i] = 1
            mono_mass = mass_list[i] - c13d2
            mass_diff = abs(mass_list - mono_mass)
            mono_pos = mass_diff == min(mass_diff)
            if min(mass_diff)/mass_list[i] * 1e6 < tol:
                if peak_intensities[mono_pos]/peak_intensities[i] < 10:
                    flags[mono_pos] = 2
    spectralFilter = flags == 0 
    filter_mass = mass_list[spectralFilter]
    filter_formula = formula_list[spectralFilter]
    filter_peak_intensities = peak_intensities[spectralFilter]
    if verbose == True:
        print(f"{len(mass_list)-len(filter_mass)} interferences removed.")
    filter_formula = list(filter_formula) #reconvert filter_formula to list to sort out data type issue
    return msTuple(filter_formula, filter_peak_intensities, filter_mass)


from .multi_van_krevelen_plot import multi_van_krevelen_plot
from .van_krevelen_histogram import van_krevelen_histogram 
from .van_krevelen_plot import van_krevelen_plot 
from .kendrick_mass_defect_plot import kendrick_mass_defect_plot
from .atomic_class_plot import atomic_class_plot
from .compound_class_plot import compound_class_plot
from .mass_histogram import mass_histogram
from .mass_spectrum import mass_spectrum
from .reaction_network import reaction_network


def element_counts(msTuple):
    """ 
	Docstring for function pykrev.element_counts
	====================
	This function takes an msTuple and gives atomic counts for C,H,N,O,P & S for each formula in the formula list.
    
	Use
	----
	element_counts(Y)
    
	Returns a list of len(Y) in which each element, i , is a dictionary containing the atomic counts of Y[0][i]. 
    
	Parameters
	----------
	Y: msTuple OR a list of elemental formula strings. 
    
    Info
    ----------
    For formula strings:
        All integers must be standard script (e.g. C6H8O7). 
        Formula strings should not contain isotopologues (e.g. C9H12O6 13C1) 
        Formula strings should only contain C,H,N,O,P and S atoms.
    """
    #Tests
    if type(msTuple) == list:
        formula_list = msTuple
    else:
        formula_list = msTuple[0]
    #Setup
    elements=['C','H','N','O','P','S']
    count_list = []
    #Main
    for formula in formula_list:
        element_numbers = dict()
        for element in elements:
            element_numbers[element] = 0
            alpha_idx = formula.find(element) #find this first instance of the element name. If the formula is given with alphabetically similar two digit element names e.g. Co before single digit names e.g. C. 
                                           #it won't work. In theory this shouldn't happen because formulas should be listed alphabetically, therefore single character names will come before two character names. 
            if alpha_idx > -1: # if the element was found in the formula   
                element_numbers[element] = 1 #it must have at least 1 atom 
                for i in range(alpha_idx+2,(len(formula)+1)):
                    try:
                        element_numbers[element] = int(formula[alpha_idx+1:i]) #Count the number of atoms belonging to this element starting with the next character, and progressively making the character window bigger
                    except ValueError: # this occurs when you have reached the next element name
                        break
        count_list.append(element_numbers)
    return count_list


import pandas as pd
import numpy as np
from .msTuple import msTuple
def read_csv(Y, column_headers = False):
    """ 
    Docstring for function pykrev.read_csv
    ==========
    Reads an assigned mass list from a .csv file into an msTuple. 
    
    Use 
    ----------
    read_csv(Y)
    
    Returns an msTuple. 
    
    Parameters 
    ----------
    Y:  String, path to the .csv file. 
        The .csv file should have three columns in the following order: formula, intensity and mass.
    
    column_headers: Bool, does the .csv file include column headers?
    """
    if column_headers == False:
        data = pd.read_csv(Y, header = None)
    else:
        data = pd.read_csv(Y)
    msTupleObj = msTuple(list(data.iloc[:,0]),np.array(data.iloc[:,1]),np.array(data.iloc[:,2]))
    return msTupleObj


from matplotlib import pyplot as plt
from ..formula.element_ratios import element_ratios 
from ..formula.element_counts import element_counts
from matplotlib.patches import Rectangle
def multi_van_krevelen_plot(msTupleDict,
                            colours=[],
                            symbols=[], 
                            alphas = [],
                            edge_colours = [],
                            x_ratio = 'OC',
                            y_ratio = 'HC',
                            patch_classes = [],
                            patch_alpha = 0.3,
                            patch_colors = ['#762a83','#9970ab','#c2a5cf','#e7d4e8','#d9f0d3','#a6dba0','#5aae61','#1b7837'],
                            patch_text = True,
                            **kwargs):
    """ 
	Docstring for function pykrev.multi_van_krevelen_plot
	====================
	This function takes multiple msTuples and plots a van Krevelen diagram. 
    
	Use
	----
	multi_van_krevelen_plot(Y)
    
	Returns tuple containing figure and axes handles 
    
	Parameters
	----------
	Y: an msTupleDict
        
	colours: A list of strings providing marker colours corresponding to each msTuple in *Y. 
	See: https://matplotlib.org/2.1.1/api/_as_gen/matplotlib.pyplot.plot.html
    
	symbol: A list of scatter plot symbols to use corresponding to each msTuple in *Y. 
	See: https://matplotlib.org/3.2.1/api/markers_api.html
    
	edge_colours: a list of strings providing marker edge colours corresponding to each msTuple in *Y.
        
	alphas: list of floats providing transparency of the marker points corresponding to each msTuple in *Y.
    x_ratio: element ratio to plot on x axis, given numerator denominator e.g. 'OC'
    y_ratio: element ratio to plot on y axis, given numerator denominator e.g. 'HC'

    patch_classes: a list of the compound classes boundaries (taken from formularity software) to overlay as patches, can include:
        'lipid-like'
        'carbohydrate-like'
        'unsaturated hydrocarbons'
        'condensed aromatics'
        'lignin-like'
        'tannin-like'
        'amino sugar-like'
        'protein-like'
    patch_alpha: the transparency of the compound class  (float between 0 and 1)
    patch_colors: hex values for the colors of each class in patch_classes 
    patch_text: boolean, include text labels on the compound class patches 
    
    **kwargs: other key word arguments to pass to plt.scatter()
    """ 
    #Tests
    assert 'alpha' not in kwargs, 'provide a list of alpha values, alphas = ...'
    assert 'c' not in kwargs, 'provide a list of colour values, colours = ...'
    assert 'color' not in kwargs, 'provide a list of color values, colours = ...'
    assert 'marker' not in kwargs, 'provide a list of marker values, symbols = ...'
    assert 'edgecolors' not in kwargs, 'provide a list of edgecolors, edge_colours = ...'
    assert 'label' not in kwargs, 'provide a list of labels, group_labels = ...'
    #Setup
    ##apply colour blind safe colors taken from https://colorbrewer2.org/ 
    cols = ['#7fc97f','#beaed4','#fdc086','#74add1','#fdae61','#abd9e9','#fee090','#e0f3f8','#ffffbf']
    sybls = ['o','X','s','D','+','1','^','v','8']
    group_labels = list(msTupleDict.keys())
    if not colours:
        colours = [cols[i] for i in range(0,len(msTupleDict))]
    if not symbols: 
        symbols = [sybls[i] for i in range(0,len(msTupleDict))]
    if not alphas:
        alphas = [0.5] * len(msTupleDict)
    if not edge_colours:
        edge_colours = ['None'] * len(msTupleDict)
    assert len(msTupleDict) == len(group_labels) == len(colours) == len(symbols) == len(alphas) == len(edge_colours), 'Input variables must all be the same length'
    i = 0 
    for msTuple in msTupleDict.values(): 
        x_axis = []
        y_axis = []
        ratio_list = element_ratios(msTuple, ratios = [x_ratio,y_ratio])
        for ratios in ratio_list:
                x_axis.append(ratios[x_ratio])
                y_axis.append(ratios[y_ratio])
        plt.scatter(x_axis, y_axis, alpha=alphas[i], edgecolors=edge_colours[i],c=colours[i], marker = symbols[i], label = group_labels[i], **kwargs)
        i += 1 
    #apply grid lines 
    plt.grid(True) 
    #add on chemical class patches
    #boundaries taken from formularity software
    assert len(patch_colors) >= len(patch_classes), "Provide at least as many colors as classes"
    cindx = 0 #index for the patch colours 
    if 'lipid-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.01,1.5),0.29,0.7,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.012,2.14,'Lipid',fontsize=11,alpha=1, color = 'k')
        cindx += 1 
    if 'carbohydrate-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.7,1.5),0.4,0.8,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.702,2.24,'Carbs',fontsize=11,alpha=1, color = 'k')
        cindx += 1 
    if 'unsaturated hydrocarbons' in patch_classes:
        plt.gca().add_patch(Rectangle((0.01,0.8),0.09,0.7,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.012,1.44,'Unsat HC',fontsize=11,alpha=1, color = 'k')
        cindx += 1
    if 'condensed aromatics' in patch_classes:
        plt.gca().add_patch(Rectangle((0.01,0.2),0.09,0.6,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.012,0.74,'Con HC',fontsize=11,alpha=1, color = 'k')
        cindx += 1 
    if 'lignin-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.1,0.8),0.6,0.8,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.102,1.54,'Lignin',fontsize=11,alpha=1, color = 'k')
        cindx += 1
    if 'tannin-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.7,0.8),0.5,0.8,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.702,1.54,'Tannin',fontsize=11,alpha=1, color = 'k')
        cindx += 1
    if 'amino sugar-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.6,1.5),0.1,0.7,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.602,2.14,'AminoSugar',fontsize=11,alpha=1, color = 'k')
        cindx += 1 
    if 'protein-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.3,1.5),0.3,0.8,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.302,2.24,'Protein',fontsize=11,alpha=1, color = 'k')
        cindx += 1  
    #label axis
    plt.xlabel(f"Atomic ratio of {x_ratio[0]}/{x_ratio[1]}")
    plt.ylabel(f"Atomic ratio of {y_ratio[0]}/{y_ratio[1]}")
    fig = plt.gcf()
    ax = plt.gca()
    return fig,ax


from ..formula.calculate_mass import calculate_mass
from matplotlib import pyplot as plt
import pandas as pd
import numpy as np
import math
def mass_histogram(msTuple,
                   method = 'monoisotopic',
                   ion_charge = -1,
                   protonated = True,
                   bin_width = [],
                   summary_statistics = False,
                   density = False,
                   hist = True,
                   kde = False,
                   kde_color = 'red',
                   **kwargs):
    """ 
	Docstring for function PyKrev.mass_histogram
	====================
	This function takes an msTuple and plots a histogram of the atomic masses or mass errors calculated using method.  
    
	Use
	----
	mass_histogram(Y)
    
    Returns the fig and ax handles and a tuple containing the mean, median and standard deviation 
    
	Parameters
	----------
	Y: msTuple
	method: the method used to calculate formula mass. See pk.calculate_mass for more information. 
            one of
            'monoisotopic'
            'average'
            'nominal'
            'mz'- plot a list of mz values in msTuple 
            'me'- plot a list of mass error values (warning: only works on deprotonated analyte e.g. [M-H]-, negative mode data)
    bin_width: an integer or float specifying the size of each histogram bin
    mz_list: a list or numpy array of mz values, must be provided to use method 'mz'. 
    summary_statistics: boolean, if true print the mean, median and standard deviation on the chart
    density: boolean, report normalised freq on y axis (as opposed to counts)
    hist: boolean, plot histogram bars.
    kde: boolean, plot a kernel density estimate line using scipy.gaussian_kde (forces density to be TRUE)
    kde_color: string, color of the kde line 
    ion_charge: int, the ion charge of the formula (required for mass error calculation) see pk.calculate_mass
    protonated: bool, shell ion type (required for mass error calculation) see pk.calculate_mass
    **kwargs: key word arguments to plt.hist
    """
    #Tests
    assert method in ['monoisotopic','average','nominal','mz','me'], 'Provide a valid method. See docstring for info.'
    #Setup
    mz_list = msTuple[2]
    expected_mass = calculate_mass(msTuple, method = 'monoisotopic', ion_charge = ion_charge, protonated = protonated)
    me_list = (expected_mass - mz_list)/mz_list * 1e6
    if method == 'mz':
        assert len(mz_list) > 1, 'provide a list of mz values'
        mass = mz_list
    elif method == 'me':
        assert len(me_list) > 1, 'provide a list of mass error values'
        mass = me_list
    else: 
        mass = calculate_mass(msTuple, method = method)
    #Main
    if kde == True:
        #use scipy gaussian_kde to compute the kde on the mass list 
        from scipy import stats
        kde = stats.gaussian_kde(mass)
        xx = np.linspace(min(mass), max(mass), 1000)
        #plot the kde
        plt.plot(xx,kde(xx),color = kde_color)
        #density must be true if kde is true
        density = True
    #calculate summary statistics
    mass_mean = np.mean(mass)
    mass_median = np.median(mass)
    mass_std = np.std(mass)
    # plot the histogram
    if hist == True: 
        if bin_width:
            assert type(bin_width) == int or type(bin_width) == float, 'Provide a scalar value for bin width'
            n = math.ceil((mass.max() - mass.min())/bin_width)
            plt.hist(x=mass, bins = n, density = density, **kwargs)
        else:
            plt.hist(x=mass, density = density, **kwargs)
    plt.grid(axis='y', alpha=0.75)
    if method == 'mz':
        plt.xlabel('m/z')
    elif method == 'me':
        plt.xlabel('Mass error (ppm)')
    else:
        plt.xlabel(f"{method[0].upper()}{method[1::]} atomic mass")
    if density == True:
        plt.ylabel("Frequency")
    else:
        plt.ylabel("Counts")
    if summary_statistics:
    ## add to the upper right of the plot
        plt.annotate(f"$\mu={np.round(mass_mean,2)}$\nm = {np.round(mass_median,2)}\n$\sigma={np.round(mass_std,2)}$", xy=(0.75, 0.75), xycoords='axes fraction')
    fig = plt.gcf()
    ax = plt.gca()
    return fig, ax , (mass_mean, mass_median, mass_std)


import numpy as np
from matplotlib import pyplot as plt
from scipy.stats import gaussian_kde
from ..formula.element_ratios import element_ratios 
from matplotlib.patches import Rectangle


def van_krevelen_plot(msTuple,
                      x_ratio = 'OC',
                      y_ratio = 'HC',
                      patch_classes = [],
                      patch_alpha = 0.3,
                      patch_colors = ['#762a83','#9970ab','#c2a5cf','#e7d4e8','#d9f0d3','#a6dba0','#5aae61','#1b7837'],
                      patch_text = True,
                      **kwargs):
    
    """ 
	Docstring for function PyKrev.van_krevelen_plot
	====================
	This function takes an msTuple and plots a van Krevelen diagram. 
    
	Use
	----
	van_krevelen_plot(Y)
  
    Returns the figure and axes handles of the plot 
    
	Parameters
	----------
	Y: msTuple
	colour: A list or numpy array of floats or integers of len(Y[0]) or
        'density' : kernel density see https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html
    x_ratio: element ratio to plot on x axis, given numerator denominator e.g. 'OC'
    y_ratio: element ratio to plot on y axis, given numerator denominator e.g. 'HC'
    patch_classes: a list of the compound classes boundaries (taken from formularity software) to overlay as patches, can include:
        'lipid-like'
        'carbohydrate-like'
        'unsaturated hydrocarbons'
        'condensed aromatics'
        'lignin-like'
        'tannin-like'
        'amino sugar-like'
        'protein-like'
    patch_alpha: the transparency of the compound class  (float between 0 and 1)
    patch_colors: hex values for the colors of each class in patch_classes 
    patch_text: boolean, include text labels on the compound class patches 
    **kwargs: key word arguments for pyplot.scatter(). 

	Info
	----------
	Given a list of ratios this function will produce Van Krevelen plots as described in 
	Kim, Sunghwan, Robert W. Kramer, and Patrick G. Hatcher. 
	"Graphical method for analysis of ultrahigh-resolution broadband mass spectra of natural organic matter, 
	the van Krevelen diagram."  
	Analytical Chemistry 75.20 (2003): 5336-5344. 
    """
    #Tests
    #check that color is provided as 'c'
    assert 'color' not in kwargs, 'supply key word color as c'
    ratio_list = element_ratios(msTuple,ratios=[x_ratio,y_ratio])
    if 'c' not in kwargs: 
        kwargs['c'] = ['blue'] * len(ratio_list)
    elif isinstance(kwargs['c'],str) and kwargs['c'] == 'density':
        kwargs['c'] = kernel_density(ratio_list)
    elif len(kwargs['c']) != len(ratio_list):
        raise ValueError('colour list and ratio list must be the same length.')
    assert len(patch_colors) >= len(patch_classes), "Provide at least as many colors as classes"
    #Setup
    x_axis = []
    y_axis = []
    for ratios in ratio_list:
            x_axis.append(ratios[x_ratio])
            y_axis.append(ratios[y_ratio])
    #Main
    plt.scatter(x_axis, y_axis, **kwargs)
    #apply grid lines 
    plt.grid(True) 
    #add on chemical class patches
    #boundaries taken from formularity software
    cindx = 0 #index for the patch colours 
    if 'lipid-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.01,1.5),0.29,0.7,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.012,2.14,'Lipid',fontsize=11,alpha=1, color = 'k')
        cindx += 1 
    if 'carbohydrate-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.7,1.5),0.4,0.8,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.702,2.24,'Carbs',fontsize=11,alpha=1, color = 'k')
        cindx += 1 
    if 'unsaturated hydrocarbons' in patch_classes:
        plt.gca().add_patch(Rectangle((0.01,0.8),0.09,0.7,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.012,1.44,'Unsat HC',fontsize=11,alpha=1, color = 'k')
        cindx += 1
    if 'condensed aromatics' in patch_classes:
        plt.gca().add_patch(Rectangle((0.01,0.2),0.09,0.6,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.012,0.74,'Con HC',fontsize=11,alpha=1, color = 'k')
        cindx += 1 
    if 'lignin-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.1,0.8),0.6,0.8,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.102,1.54,'Lignin',fontsize=11,alpha=1, color = 'k')
        cindx += 1
    if 'tannin-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.7,0.8),0.5,0.8,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.702,1.54,'Tannin',fontsize=11,alpha=1, color = 'k')
        cindx += 1
    if 'amino sugar-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.6,1.5),0.1,0.7,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.602,2.14,'AminoSugar',fontsize=11,alpha=1, color = 'k')
        cindx += 1 
    if 'protein-like' in patch_classes:
        plt.gca().add_patch(Rectangle((0.3,1.5),0.3,0.8,linewidth=2,edgecolor =patch_colors[cindx],facecolor=patch_colors[cindx],alpha = patch_alpha))
        if patch_text: plt.text(0.302,2.24,'Protein',fontsize=11,alpha=1, color = 'k')
        cindx += 1
    #add on chemical reaction lines 
    #slopes are taken from Hatcher et al. (2003) Graphical method for analysis...
    #if 'hydrogenation' in plot_reactions: 
        # need to explicitly call the axes handle, and then add a rectangle specifying the (bottom left position), width, height
        #plt.plot((0.5,0.5),(2.0,0.5), "--",alpha=1,color='#d7191c')
        #plt.text(0.52,0.52,'H',fontsize=10,alpha=1,color='#d7191c')
    #if 'redox' in plot_reactions:
        #plt.plot((0.1,0.8),(1,1), "--",alpha=1,color='#fdae61')
        #plt.text(0.81,1.02,'R/O',fontsize=10,alpha=1,color='#fdae61')
    #if 'condensation' in plot_reactions:
        #plt.plot((0.2,0.8),(0.4,1.6), "--",alpha=1,color='#ffffbf')
        #plt.text(0.82,1.58,'Co',fontsize=10,alpha=1,color='#ffffbf')
    #if 'methylation' in plot_reactions: 
        #plt.plot((0.1,.6),(1.8,.8),"--",alpha = 1,color='#abd9e9')
        #plt.text(.61,.79,'Me',fontsize=10,alpha= 1,color='#abd9e9')
    #if 'carboxylation' in plot_reactions:
        #plt.plot((0.1,0.8),(2,2),"--",alpha=1,color='#2c7bb6')
        #plt.text(0.82,2.02,'Cbx',fontsize=10,alpha=1,color='#2c7bb6')
    plt.xlabel(f"Atomic ratio of {x_ratio[0]}/{x_ratio[1]}")
    plt.ylabel(f"Atomic ratio of {y_ratio[0]}/{y_ratio[1]}")
    fig = plt.gcf()
    ax = plt.gca()
    return fig, ax 

def kernel_density(ratio_list): 
    """This function computes the kernel density of a list of molecular formula 'OC' and 'HC ratios using gaussian kernels.
       It returns a list containing the corresponding density values. For information on this function see 
       https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html                        """ 
    x = []
    y = []
    for ratios in ratio_list:
        x.append(ratios['OC'])
        y.append(ratios['HC'])
    xy =  np.vstack([x,y])
    kd = gaussian_kde(xy)(xy) #calling the inner function on (xy) and then the result on (xy)
    return list(kd)


from ..diversity.compound_class import compound_class
from matplotlib import pyplot as plt
import pandas as pd
import numpy as np

def compound_class_plot(msTuple,
                        method = 'MSCC',
                        **kwargs):
    """ 
	Docstring for function PyKrev.compound_class_plot
	====================
	This function takes an mStuple and plots a bar chart of the compound classes present.  
    
	Use
	----
	compound_class_plot(Y)
    
    Returns the figure and axes handles and a tuple containing a list of compound class assignments and a dictionary containing the compound class counts
    
	Parameters
	----------
	Y: msTuple
    mass_list: a list of mz values to pass to pk.compound_class -> required for the MSCC algorithm.
	method: the element to determine the atomic class. One of: C,H,N,O,S or P.
    **kwargs: key word arguments to plt.bar
    """
    compoundClass, cclassCounts = compound_class(msTuple,method = method)
    labels = []
    values = []
    for c in cclassCounts:      
        labels.append(str(c))
    for v in cclassCounts.values(): 
        values.append(v)
    x_pos = [i for i, _ in enumerate(labels)]
    plt.bar(x_pos, values, **kwargs)
    plt.xticks(x_pos, labels, rotation = 'vertical')
    plt.xlabel("Compound class")
    plt.ylabel("Counts")
    fig = plt.gcf()
    ax = plt.gca()
    return fig, ax, (compoundClass, cclassCounts)


from .element_counts import element_counts
import numpy as np
def aromaticity_index(msTuple, index_type = 'rAI'):     
    """ 
    Docstring for function pykrev.aromaticity_index
    ==========
    This function takes an msTuple and returns the aromaticity index of each formula in the formula list.
    
    Use
    ----------
    aromaticity_index(Y)
    
    Returns a numpy.ndarray of len(Y[0]) in which each item, i ,is the aromaticity index corresponding to Y[0][i].  
    
    Parameters
    ----------
    Y: msTuple OR a list of molecular formula strings
    
    index_type: String, one of:
        -'rAI' - reformulated aromaticity index see Mendelez-Perez et al. (2016)
        -'rAImod'- reformulated aromaticity index with modified oxygen coefficient 
        -'AI' - aromaticity index see Koch and Dittmar (2006) 
        -'AImod'- aromaticity index with modified oxygen coefficient see Koch and Dittmar (2006)
    
    
    Info
    ----------
    Reformulated aromaticity index see Mendelez-Perez et al. (2016): 
    "A reformulated aromaticity index equation under consideration for non-aromatic and non-condensed aromatic cyclic carbonyl compounds."
    Aromaticity index see Koch and Dittmar (2006): From mass to structure: an aromaticity index for high-resolution mass data of natural organic matter"
    """ 
    # Tests   
    assert index_type in ['rAI','rAImod','AI','AImod'], 'supply a valid index type, read doc string for more info'
    # Setup
    count_list = element_counts(msTuple)
    AI_array = np.array([])
    # Main
    for count in count_list: 
        try:
            if index_type == 'rAImod':
                AI_counts = (1 + 1/2*((count['C']*2)-count['H']-(count['O']*1)-(count['S']*2)-count['N']-count['P']))/count['C']
            elif index_type == 'rAI':
                AI_counts = (1 + 1/2*((count['C']*2)-count['H']-(count['O']*2)-(count['S']*2)-count['N']-count['P']))/count['C']
            elif index_type == 'AImod':
                AI_counts = (1 + count['C'] - (count['H']*0.5)-(count['O']*0.5)- count['S'])/(count['C'] - (count['O']*0.5) - count['N'] - count['S'] - count['P'])
            elif index_type == 'AI':
                AI_counts = (1 + count['C'] -count['O']- count['S'] - (0.5*count['H']))/(count['C'] - count['O'] - count['N'] - count['S'] - count['P'])
        except ZeroDivisionError:
            AI_counts = np.nan
            print(f"Warning Zero Division Encountered, NaN value returned. {count}")
        AI_array = np.append(AI_array,AI_counts)
    return AI_array


from ..formula.element_ratios import element_ratios
from ..formula.aromaticity_index import aromaticity_index
from ..formula.element_counts import element_counts
import pandas as pd
import os
def compound_class(msTuple, method = 'MSCC'):
    """ 
	Docstring for function pyKrev.compound_class
	====================
	This function takes an msTuple and calculates compound class categories based on the method give in method. 
    
	Use
	----
	compound_class(Y)
    
	Returns two items. The first is a list of len(Y[0]) in which each item is the compound class associated with the ith element of Y[0]. 
    The second is a dictionary containing the compound class counts. 
    
	Parameters
	----------
	Y: msTuple 
        
	method: String, the method to use, one of:
        'MSCC' - Multi-dimensional stoiochiometric compound classification.  See: Albert Rivas-Ubach, Yina Liu, Thomas Stephen Bianchi, Nikola Tolic, Christer Jansson, and Ljiljana Paša-Tolic. (2018)
        Moving beyond the van Krevelen diagram: A new stoichiometric approach for compound classification in organisms. Analytical Chemistry, DOI: 10.1021/acs.analchem.8b00529
        'KELL' - Compound classification based on aromaticity index. 
        See: Kellerman, A., Dittmar, T., Kothawala, D. et al. Chemodiversity of dissolved organic matter in lakes driven by climate and hydrology. Nat Commun 5, 3804 (2014).
        https://doi.org/10.1038/ncomms4804
        'FORM' - Compound classification based on the formularity algorithm. 
        'KEGG_BioMol' - Match molecular formula to those listed in 'compounds with biological roles' KEGG BRITE Pathways (databases located in ./compound_data).
        'KEGG_Phyto' - Match molecular formula to those listed in the 'phytochemicals' KEGG BRITE Heirarchy (databases located in ./compound_data). 
        'KEGG_Lipid' - Match molecular formula to those listed in the 'lipids' KEGG BRITE Heirarchy (databases located in ./compound_data). 
        'KEGG_Pesticide' - Match molecular formula to those listed in the 'pesticides' KEGG BRITE Heirarchy (databases located in ./compound_data). 
        'KEGG_Toxin' - Match molecular formula to those listed in the 'toxin' KEGG BRITE Heirarchy (databases located in ./compound_data). 
        'KEGG_All' - Match molecular formula to all of the possible categories.
        'ELEM' - Compound classification based on elemental composition. 
    
    Info
    ----------
    Please refer to the KEGG website for information on BRITE heirarchies. https://www.genome.jp/kegg/brite.html 
    """
    #Setup
    count_list = element_counts(msTuple)
    mass_list = msTuple[2]
    cclassCounts = dict()
    compound_class = []
    ##Retrieve the path to the diversity directory that this script is in, needed to retrieve compound class databases
    functionPath = os.path.realpath(__file__)
    functionDirectory = os.path.dirname(functionPath)
    #Main
    if method == 'MSCC':
        ## assert that the same number of elements in mass and count list
        assert len(mass_list) == len(count_list), 'to perform MSCC you must provide a mass list'
        ## initialise dictionary that contains counts 
        cclassCounts['Lipid'] = 0
        cclassCounts['Carbohydrate'] = 0
        cclassCounts['Amino-sugar'] = 0
        cclassCounts['Protein'] = 0
        cclassCounts['Phytochemical'] = 0
        cclassCounts['Nucleotide'] = 0
        cclassCounts['Not matched'] = 0
        cclassCounts['Double matched'] = 0
        ## initialise ratio list 
        cRatios = element_ratios(msTuple,ratios = ['OC','HC','NC','PC','NP'])
        for c,m,ratios in zip(count_list,mass_list,cRatios):
            ## a list that temporarily holds assigned categories
            cClass = []
            ## nucleotides cannot be double matched
            if ratios['OC'] >= 0.5 and ratios['OC'] < 1.7 and ratios['HC'] > 1 and ratios['HC'] < 1.8 and ratios['NC'] >= 0.2 and ratios['NC'] <= 0.5 and ratios['PC'] >= 0.1 and ratios['PC'] <= 0.35 and ratios['NP'] > 0.6 and ratios['NP'] <= 5 and c['N'] >= 2 and c['P'] >= 1 and c['S'] == 0 and m > 305 and m < 523:
                compound_class.append('Nucleotide')
                cclassCounts['Nucleotide'] += 1
            else:
                if ratios['OC'] <= 0.6 and ratios['HC'] >= 1.32 and ratios['NC'] <= 0.126 and ratios['PC'] < 0.35 and ratios['NP'] <= 5:
                    cClass.append('Lipid')
                if ratios['OC'] >= 0.8 and ratios['HC'] >= 1.65 and ratios['HC'] < 2.7  and c['N'] == 0:
                    cClass.append('Carbohydrate')
                if ratios['OC'] >= 0.61 and ratios['HC'] >= 1.45 and ratios['NC'] <= 0.2 and ratios['NC'] > 0.07 and ratios['PC'] < 0.3 and ratios['NP'] <= 2 and c['O'] >= 3 and c['N'] >= 1:
                    cClass.append('Amino-sugar')
                if ratios['OC'] <= 1.15 and ratios['HC'] < 1.32 and ratios['NC'] < 0.126 and ratios['PC'] <= 0.2 and ratios['NP'] <= 3:
                    cClass.append('Phytochemical')
                if ratios['OC'] > 0.12 and ratios['OC'] <= 0.6 and ratios['HC'] > 0.9 and ratios['HC'] < 2.5 and ratios['NC'] >= 0.126 and ratios['NC'] <= 0.7  and ratios['PC'] < 0.17 and c['N'] >= 1:
                    cClass.append('Protein')
                elif ratios['OC'] > 0.6 and ratios['OC'] <= 1 and ratios['HC'] > 1.2 and ratios['HC'] < 2.5 and ratios['NC'] > 0.2 and ratios['NC'] <= 0.7 and ratios['PC'] < 0.17 and c['N'] >= 1: 
                    cClass.append('Protein')
                if len(cClass) > 1: 
                    compound_class.append(f"Double matched: {cClass[0]} {cClass[1]}")
                    cclassCounts['Double matched'] += 1
                elif len(cClass) == 0:
                    compound_class.append("Not matched")
                    cclassCounts['Not matched'] += 1
                else:
                    compound_class.append(cClass[0])
                    cclassCounts[cClass[0]] += 1 
    if method == 'KELL':
        cRatios = element_ratios(msTuple)
        aindex = aromaticity_index(msTuple, index_type = 'rAI')
        cclassCounts['Combustion-derived polycyclic aromatics'] = 0
        cclassCounts['Vascular plant-derived polyphenols'] = 0
        cclassCounts['Highly unsaturated and phenolic compounds'] = 0
        cclassCounts['Aliphatic compounds'] = 0
        for ai,ratio in zip(aindex,cRatios):
            if ai > 0.66: 
                compound_class.append('Combustion-derived polycyclic aromatics')
                cclassCounts['Combustion-derived polycyclic aromatics'] += 1 
            elif ai > 0.5 and ai <= 0.66:
                compound_class.append('Vascular plant-derived polyphenols')
                cclassCounts['Vascular plant-derived polyphenols'] += 1 
            elif ai <= 0.5 and ratio['HC'] < 1.5: 
                compound_class.append('Highly unsaturated and phenolic compounds')
                cclassCounts['Highly unsaturated and phenolic compounds'] += 1 
            elif ai <= 0.5 and ratio['HC'] >= 1.5:
                compound_class.append('Aliphatic compounds')
                cclassCounts['Aliphatic compounds'] += 1     
    if method == 'FORM':
        cRatios = element_ratios(msTuple)
        cclassCounts['Lipid-like'] = 0
        cclassCounts['Carbohydrate-like'] = 0
        cclassCounts['Unsaturated hydrocarbons'] = 0
        cclassCounts['Condensed aromatics'] = 0
        cclassCounts['Lignin-like'] = 0
        cclassCounts['Tannin-like'] = 0
        cclassCounts['Amino Sugar-like'] = 0
        cclassCounts['Protein-like'] = 0
        cclassCounts['Not assigned'] = 0
        for ratio in cRatios:
            if ratio['OC'] >= 0.01 and ratio['OC'] <= 0.3 and ratio['HC'] >= 1.5 and ratio['HC'] <= 2.2:
                compound_class.append('Lipid-like')
                cclassCounts['Lipid-like'] += 1 
            elif ratio['OC'] >= 0.7 and ratio['OC'] <= 1.1 and ratio['HC'] >= 1.5 and ratio['HC'] <= 2.3:
                compound_class.append('Carbohydrate-like')
                cclassCounts['Carbohydrate-like'] += 1 
            elif ratio['OC'] >= 0.01 and ratio['OC'] <= 0.1 and ratio['HC'] >= 0.8 and ratio['HC'] <= 1.5:
                compound_class.append('Unsaturated hydrocarbons')
                cclassCounts['Unsaturated hydrocarbons'] += 1
            elif ratio['OC'] >= 0.01 and ratio['OC'] <= 1 and ratio['HC'] >= 0.2 and ratio['HC'] <= 0.8:
                compound_class.append('Condensed aromatics')
                cclassCounts['Condensed aromatics'] += 1
            elif ratio['OC'] >= 0.1 and ratio['OC'] <= 0.7 and ratio['HC'] >= 0.8 and ratio['HC'] <= 1.6:
                compound_class.append('Lignin-like')
                cclassCounts['Lignin-like'] += 1
            elif ratio['OC'] >= 0.7 and ratio['OC'] <= 1.2 and ratio['HC'] >= 0.8 and ratio['HC'] <= 1.6:
                compound_class.append('Tannin-like')
                cclassCounts['Tannin-like'] += 1
            elif ratio['OC'] >= 0.6 and ratio['OC'] <= 0.7 and ratio['HC'] >= 1.5 and ratio['HC'] <= 2.2:
                compound_class.append('Amino Sugar-like')
                cclassCounts['Tannin-like'] += 1
            elif ratio['OC'] >= 0.3 and ratio['OC'] <= 0.6 and ratio['HC'] >= 1.5 and ratio['HC'] <= 2.3:
                compound_class.append('Protein-like')
                cclassCounts['Protein-like'] += 1
            else:
                compound_class.append('Not assigned')
                cclassCounts['Not assigned'] += 1
    if 'KEGG' in method:
        if method == 'KEGG_BioMol':
            BRITE = pd.read_csv(f"{functionDirectory}\compound_data\Brite_BioMol_DF.csv")
        elif method == 'KEGG_Lipid': 
            BRITE = pd.read_csv(f"{functionDirectory}\compound_data\Brite_Lipid_DF.csv")
        elif method == 'KEGG_Phyto':
            BRITE = pd.read_csv(f"{functionDirectory}\compound_data\Brite_Phyto_DF.csv")
        elif method == 'KEGG_Pesticide': 
            BRITE = pd.read_csv(f"{functionDirectory}\compound_data\Brite_Pesticide_DF.csv")
        elif method == 'KEGG_Toxin': 
            BRITE = pd.read_csv(f"{functionDirectory}\compound_data\Brite_Toxin_DF.csv")
        elif method == 'KEGG_All': 
            BRITE = pd.read_csv(f"{functionDirectory}\compound_data\Brite_All_DF.csv")
        else:
            print('Error: KEGG Database Method not recognised. Refer to docstring.')
        BriteFormula = BRITE['F'].to_list()
        BriteCatA = BRITE['A'].to_list()
        KEGGCats = set(BriteCatA)
        cclassCounts['Not Matched'] = 0
        for c in KEGGCats:
            cclassCounts[c] = 0
        for f in msTuple.formula:
            try:
                idx = BriteFormula.index(f)
                compound_class.append(BriteCatA[idx])
                cclassCounts[BriteCatA[idx]] += 1
            except ValueError:
                cclassCounts['Not Matched'] += 1
                compound_class.append('Not Matched')
    if 'ELEM' in method:
        for d in count_list:
            elemclass = []
            for k,v in zip(d.keys(), d.values()):
                if v > 0:
                    elemclass.append(k)
            elemclass = ''.join(elemclass)
            if elemclass in cclassCounts.keys():
                cclassCounts[elemclass] += 1
            else:
                cclassCounts[elemclass] = 1
            compound_class.append(elemclass)
    return compound_class,cclassCounts


from ..formula.element_counts import element_counts
from matplotlib import pyplot as plt
import pandas as pd
import numpy as np

def atomic_class_plot(msTuple,
                      element = 'O',
                      summary_statistics = False,
                      **kwargs):
    """ 
	Docstring for function PyKrev.atomic_class_plot
	====================
	This function takes an msTuple and plots a histogram of the atomic classes based on element.  
    
	Use
	----
	atomic_class_plot(Y)
    
    Returns the fig and ax handles and a tuple containing the mean, median and standard deviation 
    
	Parameters
	----------
	Y: A list of molecular formula strings.

	element: String, the element to determine the atomic class. One of: C,H,N,O,S or P.

    summary_statistics: boolean, if true print the mean, median and standard deviation on the chart.

    **kwargs: key word arguments to plt.hist
    """
    #Setup
    count_list = element_counts(msTuple)
    countDF = pd.DataFrame(count_list)
    atom = np.array(countDF[element])
    atom_mean = np.mean(atom)
    atom_median = np.median(atom)
    atom_std = np.std(atom)
    #Main
    plt.hist(x=atom, **kwargs)
    plt.grid(axis='y', alpha=0.75)
    plt.xlabel(f"{element} atom class")
    plt.ylabel("Counts")
    fig = plt.gcf()
    ax = plt.gca()
    if summary_statistics:
    ## add to the upper right of the plot
        plt.annotate(f"$\mu={np.round(atom_mean,2)}$\nm = {np.round(atom_median,2)}\n$\sigma={np.round(atom_std,2)}$", xy=(0.75, 0.75), xycoords='axes fraction')
    return fig, ax, (atom_mean, atom_median, atom_std) 


from ..formula.calculate_mass import calculate_mass
import networkx as nx
import numpy as np
def reaction_network(msTuple, filePath = '', fileFormat = 'none', reactionDict = {
                                            'decarboxylation': -calculate_mass(['CO2']),
                                            'methylation': calculate_mass(['CH2']),
                                            'demethylation': -calculate_mass(['CH2']),
                                            'hydrogenation': calculate_mass(['H2']),
                                            'dehydrogenation': -calculate_mass(['H2']),
                                            'hydration': calculate_mass(['H2O']),
                                            'dehydration': -calculate_mass(['H2O']),
                                            'oxidation': calculate_mass(['O']),
                                            'reduction': -calculate_mass(['O'])
                                            }, nodeAnnotations = {}, roundVal = 8):
    """ 
	Docstring for function PyKrev.reaction_network
	====================
	This function takes an msTuple and writes a directed graph format network representation to the location set by filePath using the networkx library. 
    The reaction network has molecular formula as nodes and the reactions in reactionDict as edges. 
    The nodes can be annotated by the user with custom values.
    Edges are annotated with the associated reaction name.
    
	Use
	----
	reactionGraph(Y)
    
	Returns a tuple containing (i) a networkx representation of the graph and (ii) a dictionary of reaction counts 
    
	Parameters
	----------
	Y: msTuple
    reactionDict: dictionary, containing reaction names as keys and their associated change in monoisotopic formula mass as values.
    nodeAnnotations: dictionary, containing the node annotation names as keys, and their associated values as numpy arrays. e.g. {'Peak Intensity': intensityArray)
        where len(intensityArray) == len(formulaList)
    filePath: string, directory location to write the graph file to.
    fileFormat: string, the format of the graph file to write, either 'grapml' or 'gexf' or 'none'. If 'none' no graph file is written.
    roundVal: int, number of decimal places to round to in the mass defect calculation, default is 8
    """ 
    #Tests
    assert fileFormat in ['graphml', 'gexf', 'none'], "format must be graphml or gexf"
    assert type(filePath) == str, "filePath must be provided as a string"
    formulaList = msTuple[0]
    for value in nodeAnnotations.values():
        assert len(value) == len(formulaList), "ensure the arrays in nodeAnnotations are the same length as formula list"
    #Setup
    graphNodes = []
    reactionList = list(reactionDict.keys())
    reactionCounts = reactionDict.fromkeys(reactionDict, 0)
    formulaMass = calculate_mass(formulaList)
    #Main
    for i in range(0,len(formulaList)):
        # first process the node annotations into a usable format 
        nodeDict = {}
        for key in nodeAnnotations.keys():
            nodeDict[key] = nodeAnnotations[key][i]
        graphNodes.append((formulaList[i],nodeDict)) #set up the graph nodes with nodeAnnotations 
    G = nx.DiGraph() # Use a directed graph (edges have directional information)
    G.add_nodes_from(graphNodes) #Each formula in our sample is a node
    for j in range(0,len(formulaList)):
        for reactionType in reactionList:
            matchIndex = np.nonzero(np.round(formulaMass, roundVal) == np.round((formulaMass[j] + reactionDict[reactionType]), roundVal))
            if matchIndex[0].size > 0:
                reactionCounts[reactionType] += 1
                G.add_edge(formulaList[j],formulaList[int(matchIndex[0])], reaction = reactionType) 
    print('Number of nodes in graph:')
    print(G.number_of_nodes())
    print('Number of edges in graph')
    print(G.number_of_edges())
    if fileFormat == 'graphml':
        nx.write_graphml(G, filePath)
    elif fileFormat == 'gexf':
        nx.write_gexf(G, filePath)
    return G, reactionCounts



from ..formula.element_counts import element_counts
from ..formula.element_ratios import element_ratios
from ..formula.aromaticity_index import aromaticity_index
from ..formula.double_bond_equivalent import double_bond_equivalent
from ..formula.nominal_oxidation_state import nominal_oxidation_state
from .normalise_intensity import normalise_intensity
import numpy as np
def diversity_indices (msTuple,indices = ['r','GS','SW','C','O','NOSC','DBE','rAI','HC','OC'],verbose = True):
    """ 
	Docstring for function pyKrev.diversity_indices
	====================
	This function takes an msTuple and calculates a variety of diversity estimates (richness, abundance-based and functional.)
    
	Use
	----
	diversity_indices(Y)
    
	Returns a dictionary containing the values listed in 'indices'.   
    
	Parameters
	----------
	Y: msTuple
	indices: a list of strings specifying the specific diversity indices to calculate, can include:
        'r' : molecular richness (i.e. number of molecular formula)
        'GS': Gini-simpson abundance based alpha diversity (species eveness)
        'SW': Shannon Wiener abundance based alpha diversity
        'C' : C number functional diversity
        'O' : O number functional diversity
        'N' : N number functional diversity
        'HC': HC ratio functional diversity   
        'OC': OC ratio  functional diversity   
        'NC': NC ratio functional diversity 
        'NOSC': Nominal oxidation state of C functional diversity    
        'DBE' : Double bound equivalent functional diversity          
        'rAI' : reformulated aromaticity index functional diversity 
        'mz'  : mz values functional diversity
    verbose:  boolean, print the results to screen or not 
    
	Info
	----------
	For a greater indepth description of 
	diversity indices on molecular formula see: 'Mentges, Andrea, et al. "Functional molecular diversity of marine dissolved 
	organic matter is reduced during degradation." 
	Frontiers in Marine Science 4 (2017): 194.' 
    Functional diversity is calculated using Rao's quadratic entropy. See: "Rao, R. (1982). Diversity and dissimilarity coefficients: a unified approach." 
        
    """      
    #Setup
    formula_list = msTuple[0]
    mz_list = msTuple[2]
    intensity_list = msTuple[1]
    formula_set = set(formula_list)
    if len(formula_set) != len(formula_list):
        print('Warning: duplicates detected in formula list. Remove to avoid inaccuracies.')
    if 'mz' in indices: 
        assert len(mz_list) == len(formula_list), 'you must provide an mz list if to calculate mz functional diversity'
    #calculate element counts, element ratios and other feature types for functional diversity metric
    count_list = element_counts(formula_list)
    ratio_list = element_ratios(formula_list, ratios = ['HC','OC','NC'])
    C_list = []
    O_list = []
    N_list = []
    HC_list = []
    OC_list = []
    NC_list = []
    for i,z in zip(count_list,ratio_list):
        C_list.append(i['C'])
        HC_list.append(z['HC'])
        O_list.append(i['O'])
        OC_list.append(z['OC'])
        N_list.append(i['N'])
        NC_list.append(z['NC'])
    C_list = np.array(C_list)
    O_list = np.array(O_list)
    HC_list = np.array(HC_list)
    OC_list = np.array(OC_list)
    N_list = np.array(N_list)
    NC_list = np.array(NC_list)
    AI_list = aromaticity_index(formula_list,index_type='rAI')
    DBE_list = double_bond_equivalent(formula_list)
    NOSC_s = nominal_oxidation_state(formula_list)
    diversity_indices = dict()
    # Main
    #Firstly normalise the intensity_list
    rel_abundance = np.array(normalise_intensity(intensity_list))
    #Molecular richness is just the count of molecular formula
    if 'r' in indices: 
        D_r = len([ i for i in rel_abundance if i!= 0])
        if verbose == True:
            print('Molecular richness:',D_r,'\n')
        diversity_indices['D_r'] = D_r
    #Abundance based diversity 
    if verbose == True:
        print('Abundanced based diversity:')
    #Gini-Simpson index decribes the distribution of compounds across molecular formulas
    #The Gini simpson index ranges from 0 to 1, where larger indiate higher diversity with a maximum value of 1 - 1/N for a uniform distribution. 
    if 'GS' in indices: 
        D_a_GS = 1 - sum(rel_abundance**2)
        if verbose == True:
            print('Gini-Simpson Index:', D_a_GS)
        diversity_indices['D_a_GS'] = D_a_GS
    #The Shannon Wiener Index accounts for the evenness of individual MFs in each sample.
    if 'SW' in indices: 
        D_a_SW = -sum(rel_abundance * np.log(rel_abundance,where=(rel_abundance!=0))) #where is useful because it stops us computing the natural log of 0 which is undefined
        if verbose == True:
            print('Shannon-Wiener Index:', D_a_SW,'\n')
        diversity_indices['D_a_SW'] = D_a_SW
    #Rao's entropy uses 'absolute difference' of a chemical characteristic which can be defined as |x - y| or abs(x - y)
    if verbose == True:
        print('Functional based diversity:')
    #Calculate functional diversity based on C number, H/C ratio and oxidation state of C 
    #Include modified aromaticity index 
    D_f_C = 0
    D_f_O = 0
    D_f_N = 0
    D_f_HC = 0 
    D_f_OC = 0
    D_f_NC = 0
    D_f_rAI = 0 
    D_f_DBE = 0 
    D_f_NOSC = 0 
    D_f_mz = 0 
    for i in range(0,len(formula_list)-1):
        if 'C' in indices: 
            D_f_C += sum((rel_abundance[i] * rel_abundance[i+1::] * abs(C_list[i] - C_list[i+1::])))
        if 'O' in indices: 
            D_f_O += sum((rel_abundance[i] * rel_abundance[i+1::] * abs(O_list[i] - O_list[i+1::])))
        if 'N' in indices: 
            D_f_N += sum(rel_abundance[i] * rel_abundance[i+1::] * abs(N_list[i] - N_list[i+1::]))
        if 'HC' in indices: 
            D_f_HC += sum(rel_abundance[i] * rel_abundance[i+1::] * abs(HC_list[i] - HC_list[i+1::]))
        if 'OC' in indices: 
            D_f_OC += sum(rel_abundance[i] * rel_abundance[i+1::] * abs(OC_list[i] - OC_list[i+1::]))
        if 'NC' in indices: 
            D_f_NC += sum(rel_abundance[i] * rel_abundance[i+1::] * abs(NC_list[i] - NC_list[i+1::]))
        if 'rAI' in indices:
            D_f_rAI += sum(rel_abundance[i] * rel_abundance[i+1::] * abs(AI_list[i] - AI_list[i+1::]))
        if 'DBE' in indices:
            D_f_DBE += sum(rel_abundance[i] * rel_abundance[i+1::] * abs(DBE_list[i] - DBE_list[i+1::]))
        if 'NOSC' in indices: 
            D_f_NOSC += sum(rel_abundance[i] * rel_abundance[i+1::] * abs(NOSC_s[i] - NOSC_s[i+1::]))
        if 'mz' in indices: 
            D_f_mz += sum(rel_abundance[i] * rel_abundance[i+1::] * abs(mz_list[i] - mz_list[i+1::]))
    if 'C' in indices: 
        if verbose == True:
            print('Raos Quadratic Index (C Number): ', D_f_C)
        diversity_indices['D_f_C'] = D_f_C
    if 'O' in indices: 
        if verbose == True:
            print('Raos Quadratic Index (O Number): ', D_f_O)
        diversity_indices['D_f_O'] = D_f_O
    if 'N' in indices: 
        if verbose == True:
            print('Raos Quadratic Index (N Number): ', D_f_N)
        diversity_indices['D_f_N'] = D_f_N
    if 'HC' in indices: 
        if verbose == True:
            print('Raos Quadratic Index (HC Ratio): ', D_f_HC)
        diversity_indices['D_f_HC'] = D_f_HC
    if 'OC' in indices:
        if verbose == True:
            print('Raos Quadratic Index (OC Ratio): ', D_f_OC)
        diversity_indices['D_f_OC'] = D_f_OC
    if 'NC' in indices: 
        if verbose == True:
            print('Raos Quadratic Index (NC Ratio): ', D_f_NC)
        diversity_indices['D_f_NC'] = D_f_NC
    if 'NOSC' in indices: 
        if verbose == True:
            print('Raos Quadratic Index (NOSC): ', D_f_NOSC)
        diversity_indices['D_f_NOSC'] = D_f_NOSC
    if 'rAI' in indices: 
        if verbose == True:
            print('Raos Quadratic Index (rAI): ', D_f_rAI)
        diversity_indices['D_f_rAI'] = D_f_rAI
    if 'DBE' in indices: 
        if verbose == True:
            print('Raos Quadratic Index (DBE): ', D_f_DBE)
        diversity_indices['D_f_DBE'] = D_f_DBE
    if 'mz' in indices: 
        if verbose == True:
            print('Raos Quadratic Index (mz): ', D_f_mz)
        diversity_indices['D_f_mz'] = D_f_mz
    return diversity_indices


import numpy as np
from .element_counts import element_counts
def calculate_mass(msTuple, method = 'monoisotopic', protonated = False, ion_charge = 0):     
    """ 
	Docstring for function pyKrev.calculate_mass
	==========
	This function takes an msTuple and calculates the monoisotopic, nominal or average mass of each formula in the formula list.
    
	Use
	----------
	calculate_mass(Y)
    
	Returns a numpy array of len(Y[0]) in which each item , i , is the calculated mass of Y[0][i].
    
	Parameters
	----------
	Y: msTuple OR a list of molecular formula strings
    
    Method: String, the type of mass calculation to perform. One of:
        - 'monoisotopic', i.e. based on the exact mass of the most abundant isotope. 
        - 'nominal', i.e. the mass of the most abundant isotope rounded to the nearest integer.
        - 'average', i.e. mass taking into account weighted abundance of all natural isotopes of the element.

    protonated: Boolean, if True calculate mass for close shell ions([M + H]+  or [M - H]-) depending on ion_charge

    ion_charge: int, the ion charge of the formula, if not charged == 0 

	Info
	----------
    average masses taken from: http://physics.nist.gov/PhysRefData/Compositions/.
    monoisotopic masses taken from CoreMS python library.
    """
    #Setup 
    count_list = element_counts(msTuple)
    electron_mass = 0.0005485
    if method == 'monoisotopic':
        element_masses = {'C': 12.0, 'H': 1.007825032239, 'O':15.9949146195717,'N': 14.003074004432,'S': 31.972071174414, 'P': 30.973761998427}
    elif method == 'nominal':
        element_masses = {'C': 12, 'H': 1, 'O':16,'N': 14,'S': 32, 'P': 31}
    elif method == 'average':
        element_masses = {'C': 12.010736, 'H': 1.007941, 'O':15.999405,'N': 14.006743,'S': 32.066085, 'P': 30.973762}
    else:
        raise Exception("Method not recognised.")
    mass_list = np.empty(len(count_list))
    #Main
    for i in range(0,len(count_list)):
        mass_list[i] = count_list[i]['C'] * element_masses['C'] + count_list[i]['H'] * element_masses['H'] + count_list[i]['N'] * element_masses['N'] + count_list[i]['O'] * element_masses['O'] + count_list[i]['P'] * element_masses['P'] + count_list[i]['S'] * element_masses['S']
        if ion_charge != 0: 
            mass_list[i] += (ion_charge * -1 * electron_mass)
            if protonated == True:
                mass_list[i] += (ion_charge * element_masses['H'])
            mass_list[i] = mass_list[i]/abs(ion_charge)
    return mass_list